from copy import deepcopy
import math
from typing import (
    Any,
    Iterator,
    TypeVar,
    Annotated,
    Type,
    Union,
)
from pydantic import Tag
import pydot

from accelforge.util._basetypes import (
    EvalableModel,
    EvalableList,
    _PostCall,
    _get_tag,
)

from accelforge.util.exceptions import EvaluationError

from accelforge.frontend.arch.spatialable import Spatialable

from pydantic import Discriminator
from accelforge.util._basetypes import _uninstantiable
from accelforge.util.parallel import _SVGJupyterRender
from accelforge.util._visualization import _pydot_graph

_FIND_SENTINEL = object()

D = TypeVar("D")
T = TypeVar("T")


class ArchNode(EvalableModel):
    """A node in the architecture."""

    def find(self, name: str, default: D = _FIND_SENTINEL) -> Union["ArchNode", D]:
        """

        Finds a node with the given name.

        Parameters
        ----------
        name: str
            The name of the node to find.
        default: Any
            The value to return if a node with the given name is not found.
            Otherwise, raises a ValueError.

        Raises
        ------
        ValueError
            If a node with the given name is not found.

        Returns
        -------
        ArchNode
            The node with the given name.
        default
            The value to return if the `Leaf` node with the given name is not found.
        """
        if getattr(self, "name", None) == name:
            return self

        if isinstance(self, Branch):
            for element in self.nodes:
                try:
                    return element.find(name)
                except (AttributeError, ValueError):
                    pass
        if default is not _FIND_SENTINEL:
            return default
        raise ValueError(f"Leaf {name} not found in {self}")

    def is_above(self, node_a: str, node_b: str) -> bool:
        """Returns whether node_a is above node_b in a hierarchy."""
        self.find(node_a)
        self.find(node_b)
        for node, parents in self.iterate_hierarchically():
            if node.name != node_b:
                continue
            return any(p.name == node_a for p in parents)

    def find_first_of_type_above(
        self, node_type: T, name: str, default: D = _FIND_SENTINEL
    ) -> T | D:
        """
        Returns the first node with type `node_type` above `name`.

        If `name` does not exist, raises an error.

        If no node of `node_type` is found, either `default` is
        returned (if provided) or raises an error.
        """
        # Check if name exists
        # This *should* raise even if default != _FIND_SENTINEL
        self.find(name)

        for node, parents in self.iterate_hierarchically():
            if node.name != name:
                continue
            for p in reversed(parents):
                if isinstance(p, node_type):
                    return p
            if default is not _FIND_SENTINEL:
                return default
            raise ValueError(f"Parent of {name} with type {node_type} not found")
        raise RuntimeError(
            "BUG: find() finds node but iterate_hierarchically() does not"
        )

    def iterate_hierarchically(self, _parents=None):
        """
        Iterates over all nodes with names while also yielding the list of all
        nodes that are hierarchical parents over the current node.
        """
        if _parents is None:
            _parents = []

        if hasattr(self, "name"):
            yield self, _parents
            _parents.append(self)

        # Fork -> don't update the _parents list from MY parent because we're branching
        # off
        if isinstance(self, Fork):
            _parents = list(_parents)

        if isinstance(self, Leaf):
            return

        # Array -> Each node is independent. Create a new parent list for each one.
        elif isinstance(self, Array):
            for node in self.nodes:
                yield from node.iterate_hierarchically(list(_parents))

        # Hierarchical -> Each node will update the _parents list to add itself and
        # create the hierarchy
        elif isinstance(self, Hierarchical):
            for node in self.nodes:
                yield from node.iterate_hierarchically(_parents)
        else:
            raise RuntimeError(f"Unhandled structure type {type(self)}")

    def _render_node_name(self) -> str:
        """The name for a Pydot node."""
        return f"{self.__class__.__name__}_{id(self)}"

    def _render_node_label(self) -> str:
        """The label for a Pydot node."""
        return self.__str__()

    def _render_node_shape(self) -> str:
        """The shape for a Pydot node."""
        return "box"

    def _render_node_color(self) -> str:
        """The color for a Pydot node."""
        return "white"

    def _render_node(self) -> pydot.Node:
        """Render this node using Pydot."""
        return pydot.Node(
            self._render_node_name(),
            label=self._render_node_label(),
            shape=self._render_node_shape(),
            style="filled",
            fillcolor=self._render_node_color(),
            margin=0.05,
        )

    def _render_make_children(self) -> list[pydot.Node]:
        """Renders this node and its children, returning them as a list."""
        return [self._render_node()]


class ArchNodes(EvalableList):
    """A list of ArchNodes."""

    def __repr__(self):
        return f"{self.__class__.__name__}({super().__repr__()})"

    def _eval_expressions(self, symbol_table: dict[str, Any], *args, **kwargs):
        class PostCallArchNode(_PostCall):
            def __call__(self, field, value, evaluated, symbol_table):
                if isinstance(evaluated, Leaf):
                    symbol_table[evaluated.name] = evaluated
                return evaluated

        for i, node in enumerate(self):
            symbol_table[i] = node

        return super()._eval_expressions(
            symbol_table, *args, **kwargs, post_calls=(PostCallArchNode(),)
        )


@_uninstantiable
class Leaf(ArchNode):
    """A leaf node in the architecture. This is an abstract class that represents any
    node that is not a `Branch`."""

    name: str
    """ The name of this `Leaf`. """

    def __str__(self) -> str:
        """String representation of the Leaf node."""
        result = self.name
        if self.spatial:
            spatial_str = ", ".join(f"{s.fanout}× {s.name}" for s in self.spatial)
            result = f"{result} [{spatial_str}]"
        return result

    def _render_node_label(self) -> str:
        return f"{self.name}" + self._spatial_str()

    def _render_make_children(self) -> list[pydot.Node]:
        return [self._render_node()]


@_uninstantiable
class Branch(ArchNode):
    nodes: ArchNodes[
        Annotated[
            Union[
                Annotated["Compute", Tag("Compute")],
                Annotated["Memory", Tag("Memory")],
                Annotated["Toll", Tag("Toll")],
                Annotated["Container", Tag("Container")],
                Annotated["Network", Tag("Network")],
                Annotated["Hierarchical", Tag("Hierarchical")],
                Annotated["Array", Tag("Array")],
                Annotated["Fork", Tag("Fork")],
            ],
            Discriminator(_get_tag),
        ]
    ] = ArchNodes()

    def get_nodes_of_type(self, types: Type[T] | tuple[Type[T], ...]) -> Iterator[T]:
        for node in self.nodes:
            if isinstance(node, types):
                yield node
            elif isinstance(node, Branch):
                yield from node.get_nodes_of_type(types)

    def _render_make_children(self) -> list[pydot.Node]:
        """Renders this node and all children, returning them as a list."""
        result = []
        for node in self.nodes:
            children = node._render_make_children()
            result.extend(child for child in children if child is not None)
        return result

    def _power_gating(
        self, compute_name, used_fanout
    ) -> tuple[dict[str, float], float]:
        from accelforge.frontend.arch.structure import Fork
        from accelforge.frontend.arch.components import Compute

        result = {}
        non_power_gated_porp = 1
        found_compute = False
        i_have_compute = self.find(compute_name, default=None) is not None
        for i, node in enumerate(self.nodes):
            if isinstance(node, Fork):
                compute_in_fork = node.find(compute_name, default=None) is not None
                found_compute |= compute_in_fork

                r, _ = node._power_gating(compute_name, used_fanout)
                r = {k: v * non_power_gated_porp for k, v in r.items()}

                if node.non_forked_power_gateable and compute_in_fork:
                    non_power_gated_porp = 0
                if node.forked_power_gateable and not compute_in_fork:
                    r = {k: 0 for k in r}
                result.update(r)

            elif isinstance(node, Hierarchical):
                found_compute |= node.find(compute_name, default=None) is not None
                r, new_porp = node._power_gating(compute_name, used_fanout)
                r = {k: v * non_power_gated_porp for k, v in r.items()}
                result.update(r)
                non_power_gated_porp *= new_porp

            elif isinstance(node, Array):
                # TODO: this is a simplistic model for power gating physical
                # units
                for s in node.spatial:
                    porp = used_fanout.get((node.name, s.name), 1) / s.fanout
                    if s.power_gateable:
                        non_power_gated_porp *= porp
                r, _ = node._power_gating(compute_name, used_fanout)
                r = {k: non_power_gated_porp for k in r}
                result.update(r)

            elif isinstance(node, Leaf):
                for s in node.spatial:
                    if found_compute or not i_have_compute:
                        assert (node.name, s.name) not in used_fanout, "BUG"
                        porp = 0
                    else:
                        porp = used_fanout.get((node.name, s.name), 1) / s.fanout
                    if s.power_gateable:
                        non_power_gated_porp *= porp
                result[node.name] = non_power_gated_porp

                if isinstance(node, Compute) and node.name == compute_name:
                    found_compute = True

            else:
                raise TypeError(f"Unknown node type: {type(node)}")

        return result, non_power_gated_porp


class Array(Branch, Spatialable):
    name: str

    def model_post_init(self, __context__=None) -> None:
        for node in self.nodes:
            if isinstance(node, Fork):
                raise EvaluationError("cannot have fork inside array")

    def _flatten(
        self,
        compute_node: str,
        fanout: int = 1,
        return_fanout: bool = False,
    ):
        from accelforge.frontend.arch.components import Compute

        nodes = []

        for node in self.nodes:
            try:
                if isinstance(node, Branch):
                    raise RuntimeError("do not put branches inside array")
                elif isinstance(node, Leaf):
                    if isinstance(node, Spatialable):
                        fanout *= node.get_fanout()
                        node = deepcopy(node)
                        node._physical_spatial = node.spatial
                        node.spatial = EvalableList()
                    nodes.append(node)
                else:
                    raise RuntimeError(f"unhandled structure type {node}")
            except EvaluationError as e:
                e.add_field(node)
                raise e

        if return_fanout:
            return nodes, fanout
        return nodes

    def _render_node_label(self) -> str:
        return f"Array {self._spatial_str()}"

    def _render_node_color(self) -> str:
        return "#FCC2FC"

    def _parent2child_names(
        self, parent_name: str = None
    ) -> tuple[list[tuple[str, str, str]], str]:
        from accelforge.frontend.arch.components import Compute

        edges = []
        current_parent_name = parent_name

        for node in self.nodes:
            if isinstance(node, Branch):
                raise EvaluationError("do not put branch in array")
            elif isinstance(node, Compute):
                # Compute nodes branch off to the side like a Fork
                if current_parent_name is not None:
                    edges.append(
                        (current_parent_name, node._render_node_name(), "dashed")
                    )
            else:
                if current_parent_name is not None:
                    edges.append(
                        (current_parent_name, node._render_node_name(), "dashed")
                    )
        return edges, self._render_node_name()

    def _render_make_children(self) -> list[pydot.Node]:
        """Renders only children, not the Hierarchical node itself."""
        result = [self._render_node()]
        for node in self.nodes:
            children = node._render_make_children()
            result.extend(child for child in children if child is not None)
        return result

    def render(self) -> str:
        """Renders the architecture as a Pydot graph."""
        graph = _pydot_graph()

        # Render all nodes (Hierarchical nodes return None and are filtered out)
        for node in self._render_make_children():
            if node is not None:
                graph.add_node(node)

        # Add edges
        edges, _ = self._parent2child_names()
        for parent_name, child_name in edges:
            graph.add_edge(pydot.Edge(parent_name, child_name))

        return _SVGJupyterRender(graph.create_svg(prog="dot").decode("utf-8"))

    def _repr_svg_(self) -> str:
        return self.render()


class Hierarchical(Branch):
    def _flatten(
        self,
        compute_node: str,
        fanout: int = 1,
        return_fanout: bool = False,
    ):
        from accelforge.frontend.arch.components import Compute

        nodes = []

        for i, node in enumerate(self.nodes):
            try:
                if isinstance(node, Hierarchical):
                    if isinstance(node, Fork):
                        # If it's a compute node and our node is not in the fork, skip
                        # it
                        if node.find(compute_node, default=None) is None:
                            continue
                    new_nodes, new_fanout = node._flatten(
                        compute_node, fanout, return_fanout=True
                    )
                    nodes.extend(new_nodes)
                    fanout *= new_fanout
                    if any(
                        isinstance(n, Compute) and n.name == compute_node
                        for n in new_nodes
                    ):
                        break
                    assert not isinstance(node, Fork)
                elif isinstance(node, Array):
                    new_nodes = node._flatten(compute_node, fanout, return_fanout=False)
                    nodes.extend(new_nodes)
                    nodes.append(node)
                    fanout *= node.get_fanout()
                    if any(
                        isinstance(n, Compute) and n.name == compute_node
                        for n in new_nodes
                    ):
                        break
                elif isinstance(node, Compute):
                    if node.name == compute_node:
                        fanout *= node.get_fanout()
                        nodes.append(node)
                        break
                elif isinstance(node, Leaf):
                    fanout *= node.get_fanout()
                    nodes.append(node)
                else:
                    raise TypeError(f"Can't flatten {node}")
            except EvaluationError as e:
                e.add_field(node)
                raise e

        if return_fanout:
            return nodes, fanout
        return nodes

    def _parent2child_names(
        self, parent_name: str = None
    ) -> tuple[list[tuple[str, str]], str]:
        from accelforge.frontend.arch.components import Compute

        edges = []
        current_parent_name = parent_name

        for node in self.nodes:
            if isinstance(node, (Hierarchical, Fork)):
                child_edges, last_child_name = node._parent2child_names(
                    current_parent_name
                )
                edges.extend(child_edges)
                if last_child_name is not None:
                    current_parent_name = last_child_name
            elif isinstance(node, Array):
                child_edges, last_child_name = node._parent2child_names(
                    node._render_node_name()
                )
                edges.extend(child_edges)
                edges.append((current_parent_name, last_child_name, "solid"))
                if last_child_name is not None:
                    current_parent_name = last_child_name
            elif isinstance(node, Compute):
                # Compute nodes branch off to the side like a Fork
                if current_parent_name is not None:
                    edges.append(
                        (current_parent_name, node._render_node_name(), "solid")
                    )
            else:
                if current_parent_name is not None:
                    edges.append(
                        (current_parent_name, node._render_node_name(), "solid")
                    )

                # Update parent for next iteration
                current_parent_name = node._render_node_name()

        return edges, current_parent_name

    def _render_make_children(self) -> list[pydot.Node]:
        """Renders only children, not the Hierarchical node itself."""
        result = []
        for node in self.nodes:
            children = node._render_make_children()
            result.extend(child for child in children if child is not None)
        return result

    def render(self) -> str:
        """Renders the architecture as a Pydot graph."""
        graph = _pydot_graph()

        # Render all nodes (Hierarchical nodes return None and are filtered out)
        for node in self._render_make_children():
            if node is not None:
                graph.add_node(node)

        # Add edges
        edges, _ = self._parent2child_names()
        for parent_name, child_name, style in edges:
            graph.add_edge(pydot.Edge(parent_name, child_name, style=style))

        return _SVGJupyterRender(graph.create_svg(prog="dot").decode("utf-8"))

    def _repr_svg_(self) -> str:
        return self.render()


class Fork(Hierarchical):
    """
    A Fork is a Hierarchical that branches off from the main path. The nodes
    inside the Fork are a separate branch, while the main path continues to the next
    sibling after the Fork.
    """

    forked_power_gateable: bool = False
    """
    Whether the child branch (the nodes inside this Fork) can be power gated when
    the main branch is active. If True, these nodes will not leak when the main
    branch is being used.
    """

    non_forked_power_gateable: bool = False
    """
    Whether the main branch (the siblings after this Fork in the parent) can be power
    gated when the child branch is active. If True, those nodes will not leak when
    this Fork's child branch is being used.
    """

    def _parent2child_names(
        self, parent_name: str = None
    ) -> tuple[list[tuple[str, str]], str]:
        edges, _ = super()._parent2child_names(parent_name)
        return edges, parent_name



"""
A module containing the visualization and types needed to run mapspace exploratioon
in AccelForge.
"""

import copy
from dataclasses import dataclass
import inspect
import itertools
from math import ceil
import pydot

from accelforge.util._frozenset import oset
from typing import (
    # Collections
    Any,
    List,
    # Object definitions
    Annotated,
    Callable,
    Literal,
    Self,
    # Type constructions
    Type,
    TypeVar,
    TypeAlias,
    # Variable meta-mandates
    Optional,
    Union,
    override,
)
from collections.abc import Set
from pydantic import ConfigDict, Discriminator, Tag, computed_field
import sympy

from accelforge.frontend.renames import EinsumName, TensorName
from accelforge.util._basetypes import (
    # Parsing helpers for the input files.
    EvalableModel,
    EvalableList,
    EvalsTo,
    # Retrieves information from YAML tags.
    _get_tag,
    _uninstantiable,
    NoParse,
)
from accelforge.frontend.workload import RankVariable, TensorName
from accelforge.util._visualization import ColorMap, _pydot_graph
from accelforge.util.parallel import _SVGJupyterRender
from accelforge.frontend import arch

T = TypeVar("T", bound="MappingNode")
"""TypeVar T: Restricts the allowable types to types of MappingNodes."""

NodeList: TypeAlias = EvalableList[
    Annotated[
        Union[
            Annotated["Split", Tag("Split")],
            Annotated["Compute", Tag("Compute")],
            Annotated["Storage", Tag("Storage")],
            Annotated["Temporal", Tag("Temporal")],
            Annotated["Spatial", Tag("Spatial")],
            Annotated["Sequential", Tag("Sequential")],
            Annotated["Pipeline", Tag("Pipeline")],
            Annotated["Nested", Tag("Nested")],
            Annotated["Reservation", Tag("Reservation")],
            Annotated["Mapping", Tag("Mapping")],
            Annotated["Toll", Tag("Toll")],
        ],
        Discriminator(_get_tag),
    ]
]
"""
TypeAlias NodeList: EvalableList that can contain and discriminate between
MappingNodes of different types.
"""

_NO_JOIN_MAPPING_VISUALIZATION = False

# =============================================================================
# LoopTree Mapping Nodes
# =============================================================================


def _as_oset(x):
    """Wrap a single rank variable as a singleton oset; pass an oset through."""
    return oset(x) if isinstance(x, (set, frozenset)) else oset([x])


@_uninstantiable
class MappingNode(EvalableModel):
    """
    Represents a Node in the Mapping, which can be a loop, a storage node, a compute
    node, etc.
    """

    _constraint_lambdas: List[Callable[[], bool]] = []
    """ Constraints that apply to this node. """

    _must_be_here: bool = False
    """ Can the mapper move this node? """

    _required: bool = False
    """ Must the mapper keep this node? """

    def _render_node_name(self) -> str:
        """The name for a Pydot node."""
        return f"{self.__class__.__name__}_{id(self)}"

    def _render_node_label(self, **kwargs) -> str:
        """The label for a Pydot node."""
        return self.__str__()

    def _render_node_shape(self) -> str:
        """The shape for a Pydot node."""
        return "box"

    def _render_node(self, **kwargs) -> str:
        """Render this node using Pydot."""
        return pydot.Node(
            self._render_node_name(),
            label=self._render_node_label(**kwargs),
            shape=self._render_node_shape(),
            style="filled",
            fillcolor=self._render_node_color(),
            margin=0.05,
        )

    def _parent2next(self) -> "MappingNode":
        """
        Return the parent to the next node in the tree. This is used for nodes that
        don't appear in the tree, like Nested nodes.
        """
        return self

    def _parent2child(
        self, parent: "MappingNode"
    ) -> list[tuple["MappingNode", "MappingNode"]]:
        """
        Returns a list of tuples, each one being a parent and child node.
        """
        return []

    def _render_make_children(self, **kwargs) -> list[str]:
        """
        Renders the children of this node and returns them as a list of strings.
        """
        return []

    def get_nodes_of_type(self, types: Type[T] | tuple[Type[T], ...]) -> List[T]:
        """
        Returns all sub-nodes, including this one, that match the given types.
        """
        nodes: List[T] = []
        if isinstance(self, types):
            nodes.append(self)
        if isinstance(self, MappingNodeWithChildren):
            for node in self.nodes:
                if isinstance(node, types):
                    nodes.append(node)
                if isinstance(node, MappingNodeWithChildren):
                    nodes.extend(node.get_nodes_of_type(types))
        return nodes

    def _flatten(self) -> list["MappingNode"]:
        if isinstance(self, MappingNodeWithChildren):
            result = [self]
            for node in self.nodes:
                result.extend(node._flatten())
            return result
        return [self]

    def _render_node_color(self) -> str:
        """The color for a Pydot node."""
        return "white"

    def __hash__(self):
        """
        Hashing functor to create mappings of nodes to other objects.
        """
        return id(self)

    def __eq__(self, other: Any):
        return self is other

    def __init_subclass__(cls, **kwargs):
        # Let Pydantic build the subclass first.
        super().__init_subclass__(**kwargs)
        # Read the *raw* attribute without descriptor binding,
        h = inspect.getattr_static(cls, "__hash__", None)
        # Replace if unhashable (None) or if it's just BaseModel’s default.
        if h is None or h is EvalableModel.__hash__:
            cls.__hash__ = MappingNode.__hash__

    def compact_str(self) -> str:
        """Returns a compact string representation of this node."""
        return self.__str__()


@dataclass(frozen=True)
class TilePattern:
    tile_shape: EvalsTo[
        Literal["symbol"] | sympy.Symbol | int | str | None | sympy.Expr
    ] = "symbol"
    """
    The common tile shape of the pattern. This is the number of indices by which
    the tile moves each iteration.
    """

    initial_tile_shape: EvalsTo[
        Literal["symbol"] | sympy.Symbol | int | None | str | sympy.Expr
    ] = "symbol"
    """
    The initial tile shape. This is the shape of the tile at the first iteration.
    Subsequent iterations may be smaller if they overlap previous iterations.
    """

    calculated_n_iterations: (
        Literal["symbol"] | sympy.Symbol | int | None | str | sympy.Expr
    ) = None
    """ The number of iterations in the pattern. Do not set this! Used internally by the
    mapper. """

    def _symbol_attrs(self) -> tuple[str, ...]:
        """The attributes that may be symbols."""
        return ("tile_shape", "initial_tile_shape", "calculated_n_iterations")

    def __str__(self) -> str:
        return self.as_str()

    def as_str(self, with_initial_tile_shape=True, with_tile_shape=True):
        s = []
        if self.calculated_n_iterations not in (None, "symbol"):
            s.append(f"in [0..{self.calculated_n_iterations})")
        if with_initial_tile_shape and (
            self.initial_tile_shape not in (None, "symbol")
        ):
            s.append(f"initial={self.initial_tile_shape}")
        if with_tile_shape and (self.tile_shape not in (None, "symbol")):
            s.append(f"tile_shape={self.tile_shape}")
        return " ".join(s)

    def update(self, **kwargs) -> "TilePattern":
        """Update the TilePattern with the given keyword arguments."""
        return type(self)(**{**self.__dict__, **kwargs})

    def _symbol2str(self) -> "TilePattern":
        """
        Convert the symbols in the TilePattern to strings, and return a new TilePattern
        with the symbols replaced by their names.
        """

        def _symbol2str(x: sympy.Symbol | int | None) -> str | int | None:
            return x.name if isinstance(x, sympy.Symbol) else x

        return type(self)(
            **{x: _symbol2str(getattr(self, x)) for x in self._symbol_attrs()}
        )

    def _prepend_symbols(self, prepend: str) -> "TilePattern":
        def _prepend(x: sympy.Symbol | int | None) -> str | int | None:
            if isinstance(x, sympy.Symbol):
                x = x.name
            return prepend + x if isinstance(x, str) else x

        return self.update(
            {x: _prepend(getattr(self, x)) for x in self._symbol_attrs()}
        )

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, TilePattern):
            return False
        return all(getattr(self, x) == getattr(other, x) for x in self._symbol_attrs())

    def __hash__(self) -> int:
        return hash((self.initial_tile_shape, self.tile_shape))

    def _rename_to_match(
        self, other: "TilePattern"
    ) -> tuple["TilePattern", dict[str, str]]:
        """
        Changes the symbols in this TilePattern to match the other TilePattern.

        Parameters
        ----------
        other:
            The TilePattern to match.

        Returns
        -------
        A tuple containing the updated TilePattern and a dictionary of source->target
        symbol renames.
        """
        renames = {}
        setattrs = {}
        for x in self._symbol_attrs():
            if getattr(self, x) != getattr(other, x):
                renames[getattr(self, x)] = getattr(other, x)
                setattrs[x] = getattr(other, x)
        return self.update(**setattrs), renames

    def _clear_symbols(self) -> "TilePattern":
        """
        Clears the symbols in this TilePattern, replacing them with None.
        """

        def desymbol(x: str | sympy.Symbol | int | None) -> str | int | None:
            if isinstance(x, (str, sympy.Symbol)):
                return None
            return x

        return self.update(
            **{x: desymbol(getattr(self, x)) for x in self._symbol_attrs()}
        )


@_uninstantiable
class Loop(MappingNode):
    """
    A bounded loop over a rank variable with a given shape and/or pattern.

    Do not instantiate directly; inherited by :class:`~.Temporal` and
    :class:`~.Spatial`.
    """

    rank_variable: set[RankVariable] | RankVariable
    """ The rank variable(s) iterated over in this loop. This may be a
    single rank variable, or a set of rank variables if the loop is shared between
    multiple Einsums.

    NOTE FOR DEVELOPERS: If the rank variable is a set DURING PMAPPIN GENERATION, then
    it means that the loop is a placeholder for multiple loops in the given Einsum. They
    will be split into multiple loops later. If the rank variable

    """

    tile_shape: EvalsTo[sympy.Symbol | sympy.Expr | int | str] = "symbol"
    """
    The (common) tile shape of the iteration. For example, if the iteration
    space is range(6) and the tile shape is 3, then we create and iterate over
    two tiles [0, 1, 2] and [3, 4, 5].

    This attribute specifies the *common* tile shape because
    `initial_tile_shape` may be specified.

    For users writing YAML, the value should be an integer.

    For those developing the mapper, the literal string "symbol" is often used
    to tell the model to create a sympy symbol to use as the tile shape. Any
    other string may be specified to explicitly request a variable name (later
    converted to a sympy variable).
    """

    initial_tile_shape: EvalsTo[sympy.Symbol | sympy.Expr | int | str | None] = None
    """
    The shape of the first tile shape. This attribute is optional. If not
    specified, all tiles have the same shape.

    If specified, the initial tile shape may differ. For example, an initial
    tile shape of 3 and tile shape of 2 creates the following tiles in the
    iteration space: [0, 1, 2], [3, 4], [5, 6], ...

    Similarly to tile shape, this value should be an integer when writing a
    YAML input.

    For those developing the mapper, this attribute can be a string. See
    tile_shape for details.
    """

    _einsum_to_rank_variable: dict[EinsumName, RankVariable] = {}
    """
    If this loop has multiple rank variables (because it was merged from loops
    of different Einsums), this dictionary records which rank variable each
    Einsum originally used for this loop.
    """

    _calculated_n_iterations: (
        Literal["symbol"] | sympy.Symbol | sympy.Expr | int | str | None
    ) = None

    _fused: bool = None
    """ Whether this Loop is shared with another Einsum. """

    _may_cause_imperfect: bool = False
    """
    This means that the tile shape of this loop may not perfectly factorize the rank
    shape.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __str__(self) -> str:
        return f"for {self.rank_variable} {self.tile_pattern}"

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, Loop)
            and self.rank_variable == other.rank_variable
            and self.tile_pattern == other.tile_pattern
        )

    def _render_node_shape(self) -> str:
        return "box"

    def _render_node_color(self) -> str:
        return "#FCC2FC"

    @override
    def compact_str(self) -> str:
        """Returns a compact string representation of this Loop."""
        rv = self.rank_variable
        if isinstance(rv, (set, frozenset)):
            rv = ",".join(sorted(rv))
        return f"{rv} {self.tile_pattern}"

    def _merge(self, other: "Loop", **kwargs) -> "Loop":
        """Merge this Loop with another Loop, returning the result."""
        if not isinstance(other, Loop):
            raise ValueError(f"Expected Loop, got {type(other)}")
        if self.tile_pattern != other.tile_pattern:
            raise ValueError(
                f"Tile patterns do not match: {self.tile_pattern} != {other.tile_pattern}"
            )

        my_rv, other_rv = self.rank_variable, other.rank_variable
        my_rv = my_rv if isinstance(my_rv, (set, frozenset)) else oset((my_rv,))
        other_rv = (
            other_rv if isinstance(other_rv, (set, frozenset)) else oset((other_rv,))
        )
        return type(self)(
            rank_variable=my_rv | other_rv,
            tile_pattern=self.tile_pattern,
            **kwargs,
        )

    @property
    def tile_pattern(self) -> TilePattern:
        return TilePattern(
            tile_shape=self.tile_shape,
            initial_tile_shape=self.initial_tile_shape,
            calculated_n_iterations=self.calculated_n_iterations,
        )

    @tile_pattern.setter
    def tile_pattern(self, value: TilePattern):
        self.tile_shape = value.tile_shape
        self.initial_tile_shape = value.initial_tile_shape
        self.calculated_n_iterations = value.calculated_n_iterations

    @property
    def calculated_n_iterations(self) -> int:
        """The number of iterations performed by this loop."""
        return self._calculated_n_iterations

    @calculated_n_iterations.setter
    def calculated_n_iterations(self, value: int) -> None:
        """Set the number of iterations performed by this loop. Do not set this!
        This is calculated by the Mapper."""
        self._calculated_n_iterations = value


class Temporal(Loop):
    """A Temporal :class:`~.Loop`."""

    @override
    def compact_str(self) -> str:
        return f"T-{super().compact_str()}"

    def __eq__(self, other: "Temporal") -> bool:
        return isinstance(other, Temporal) and super().__eq__(other)

    def _merge(self, other: "Temporal") -> "Temporal":
        if not isinstance(other, Temporal):
            raise ValueError(f"Expected Temporal, got {type(other)}")
        return super()._merge(other)

    def _render_node_label(self, **kwargs) -> str:
        with_initial_tile_shape = True
        with_tile_shape = kwargs.get("with_tile_shape", True)
        return (
            f"for {self.rank_variable} "
            f"{self.tile_pattern.as_str(with_initial_tile_shape, with_tile_shape)}"
        )


class Spatial(Loop):
    """A spatial :class:`~.Loop`."""

    name: int | str
    """ The dimension over which the spatial is occuring. """

    component: str
    """ The component name across which different spatial iterations occur. """

    component_object: NoParse[arch.ArchNode] = None
    """ The component object across which different spatial iterations occur. """

    _constrained_to_one: bool = False
    """ Whether this Spatial loop is constrained to one iteration. Do not set this; used
    internally by the Mapper."""

    @override
    def compact_str(self) -> str:
        return f"S-{self.name}-{super().compact_str()}"

    def __str__(self) -> str:
        return f"S-{self.name} " + super().__str__()

    def __eq__(self, other: "Spatial") -> bool:
        return (
            isinstance(other, Spatial)
            and super().__eq__(other)
            and self.name == other.name
            and self.component == other.component
        )

    def _merge(self, other: "Spatial") -> "Spatial":
        if not isinstance(other, Spatial):
            raise ValueError(f"Expected Spatial, got {type(other)}")
        if self.name != other.name:
            raise ValueError(f"Names do not match: {self.name} != {other.name}")
        if self.component != other.component:
            raise ValueError(
                f"Components do not match: {self.component} != {other.component}"
            )
        return super()._merge(
            other,
            name=self.name,
            component=self.component,
            component_object=self.component_object,
        )

    def _render_node_label(self, **kwargs) -> str:
        with_initial_tile_shape = kwargs.get("with_initial_tile_shape", True)
        with_tile_shape = kwargs.get("with_tile_shape", True)
        return (
            f"S-{self.name}-for {self.rank_variable} "
            f"{self.tile_pattern.as_str(with_initial_tile_shape, with_tile_shape)}"
        )


class TensorHolder(MappingNode):
    """A node that represents a hardware Component holding a set of tensors."""

    tensors: EvalableList[TensorName]
    """ The names of the tensors being held in this node. """

    component: str
    """ The name of the component holding the tensors. """

    component_object: NoParse[arch.Component] = None
    """ The component object holding the tensors. """

    _must_keep_tensors: EvalableList[TensorName] = EvalableList()
    """ Which tensor(s) the Mapper must keep here. Do not set this! Used internally by
    the Mapper."""

    _backing: Set[TensorName] = oset()
    """ Which tensor(s) are backed by this node. Do not set this! Used internally by
    the Mapper."""

    _lower: bool = True
    """ Whether this tensor holder can be lowered. Do not set this! Used internally by
    the Mapper."""

    persistent: bool = False
    """
    Whether this tensor holder is persistent. Persistent tensors can't be tiled and must
    be kept in backing storage for the full duration of the workload's execution.
    """

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, TensorHolder)
            and oset(self.tensors) == oset(other.tensors)
            and self.component == other.component
        )

    @override
    def compact_str(self) -> str:
        tname = ",".join(self.tensors)
        return f"[{tname} in {self.component}]"

    def __str__(self, color_map: ColorMap = None) -> str:
        tensors = self.tensors
        if color_map is not None:
            format_list = [f"{self.component} reuses"] + list(tensors)
            return color_map.format_list(format_list)
        return f"{self.component} reuses {', '.join(tensors)}"

    @property
    def tensor(self) -> TensorName:
        """If there is one tensor held in this tensor holder, returns its name.
        Otherwise, raises an error."""
        if len(self.tensors) != 1:
            raise ValueError(
                f"TensorHolder node {repr(self)} has {len(self.tensors)} tensors. "
                f"Access the tensors property instead."
            )
        return self.tensors[0]

    def _render_node_shape(self) -> str:
        return "cylinder"

    def _render_node_color(self) -> str:
        return "#D7FCD7"

    def _merge(self, other: "TensorHolder") -> "TensorHolder":
        if not isinstance(other, TensorHolder):
            raise ValueError(f"Expected TensorHolder, got {type(other)}")

        if self.component != other.component:
            raise ValueError(
                f"Components do not match: {self.component} != {other.component}"
            )

        if self._lower != other._lower:
            raise ValueError(
                f"Lower flags do not match: {self._lower} != {other._lower}"
            )

        new = type(self)(
            tensors=self.tensors + other.tensors,
            component=self.component,
            component_object=self.component_object,
            _lower=self._lower,
        )
        return new


class Storage(TensorHolder):
    """
    A Storage :class:`~.TensorHolder` that can hold tensors for reuse.
    """

    def _merge(self, other: "Storage") -> "Storage":
        if not isinstance(other, Storage):
            raise ValueError(f"Expected Storage, got {type(other)}")
        return super()._merge(other)


class Toll(TensorHolder):
    """
    A Toll :class:`~.TensorHolder` that acts as a pass-through, where data is
    not reused but incurs accesses into this Toll.
    """

    def _render_node_shape(self) -> str:
        return "rarrow"

    def _render_node_color(self) -> str:
        return "#FFCC99"

    def __str__(self, color_map: ColorMap = None) -> str:
        tensors = self.tensors
        if color_map is not None:
            format_list = [f"{self.component} passes"] + list(tensors)
            return color_map.format_list(format_list)
        return f"{self.component} passes {', '.join(tensors)}"


class Compute(MappingNode):
    """A node that represents a compute operation. These nodes are the leaves of the
    LoopTree."""

    einsum: str
    """ The Einsum being computed. """

    component: str
    """ The name of the compute component performing the computation. """

    component_object: NoParse[arch.Compute | None] = None
    """ The :class:`~accelforge.frontend.arch.Compute` object performing the
    computation. """

    @override
    def compact_str(self) -> str:
        return f"{self.component} computes {self.einsum}"

    def __str__(self) -> str:
        return f"{self.component} computes {self.einsum}"

    def _render_node_shape(self) -> str:
        return "ellipse"

    def _render_node_color(self) -> str:
        return "#E0EEFF"


class MappingNodeWithChildren(MappingNode):
    """
    A :class:`~.MappingNode` that also has child nodes.
    """

    nodes: NodeList = EvalableList()
    """ The child nodes. """

    @override
    def _parent2child(
        self, parent: MappingNode
    ) -> list[tuple[MappingNode, MappingNode]]:
        mine = [(self, node) for node in self.nodes]
        for child in self.nodes:
            mine.extend(child._parent2child(self))
        return mine

    @override
    def _parent2next(self) -> MappingNode:
        return None

    @override
    def _render_make_children(self, **kwargs) -> list[str]:
        exclude_types = kwargs.get("exclude_types", tuple())
        lines = []
        for child in self.nodes:
            if not isinstance(child, exclude_types):
                lines.append(child._render_node(**kwargs))
            lines.extend(child._render_make_children(**kwargs))
        return lines

    @override
    def _get_backers(self) -> list[TensorHolder]:
        backing = []
        for child in self.nodes:
            if isinstance(child, TensorHolder) and child._backing:
                backing.append(child)
            elif isinstance(child, MappingNodeWithChildren):
                backing.extend(child._get_backers())
        return backing

    def clear_nodes_of_type(self, types: type | tuple[type]) -> None:
        """Clears all child nodes that match the given type(s)."""
        new_nodes = []
        for node in self.nodes:
            if isinstance(node, types):
                continue
            if isinstance(node, MappingNodeWithChildren):
                node.clear_nodes_of_type(types)
            new_nodes.append(node)
        self.nodes = EvalableList(new_nodes)

    def clear_nodes(self, *nodes: MappingNode) -> None:
        """Removes nodes that equal any of the given nodes."""
        new_nodes: list[MappingNode] = []
        for node in self.nodes:
            if any(n == node for n in nodes):
                continue
            if node in nodes:
                continue
            if isinstance(node, MappingNodeWithChildren):
                node.clear_nodes(*nodes)
            new_nodes.append(node)
        self.nodes = EvalableList(new_nodes)

    def _consolidate_tensor_holders(self) -> None:
        new_nodes = []
        for node in self.nodes:
            if isinstance(node, TensorHolder):
                # Remove duplicates
                seen_loop = False
                node = copy.deepcopy(node)
                for n in new_nodes[::-1]:
                    if isinstance(n, TensorHolder) and n.component == node.component:
                        drop_tensors = [t for t in node.tensors if t in n.tensors]
                        assert not (drop_tensors and seen_loop), "BUG"
                        node.tensors = [t for t in node.tensors if t not in n.tensors]
                    if isinstance(n, Loop):
                        seen_loop = True

                found = False
                for n in new_nodes[::-1]:
                    if isinstance(n, TensorHolder) and n.component == node.component:
                        n.tensors.extend(
                            n2 for n2 in node.tensors if n2 not in n.tensors
                        )
                        found = True
                        break
                    if isinstance(n, Loop):
                        break
                    # Don't lift a storage above a storage for the same tensor
                    if isinstance(n, TensorHolder) and oset(n.tensors) & oset(
                        node.tensors
                    ):
                        break
                    # Don't lift a storage above a reservation for the same tensor & a
                    # different component
                    if (
                        isinstance(n, Reservation)
                        and oset(n.purposes) & oset(node.tensors)
                        and n.resource != node.component
                    ):
                        break
                if not found and node.tensors:
                    new_nodes.append(node)
            else:
                new_nodes.append(node)
            if isinstance(node, MappingNodeWithChildren):
                node._consolidate_tensor_holders()
        assert new_nodes, "BUG"
        self.nodes = EvalableList(new_nodes)

    def _consolidate_reservations(self) -> None:
        new_nodes = []
        for node in self.nodes:
            if isinstance(node, Reservation):
                # Remove duplicates
                seen_loop = False
                node = copy.deepcopy(node)
                for n in new_nodes[::-1]:
                    if isinstance(n, Reservation) and n.resource == node.resource:
                        drop_purposes = [p for p in node.purposes if p in n.purposes]
                        assert not (drop_purposes and seen_loop), "BUG"
                        node.purposes = [
                            p for p in node.purposes if p not in n.purposes
                        ]
                    if isinstance(n, Loop):
                        seen_loop = True

                found = False
                for n in new_nodes[::-1]:
                    if isinstance(n, Reservation) and n.resource == node.resource:
                        n.purposes.extend(
                            n2 for n2 in node.purposes if n2 not in n.purposes
                        )
                        found = True
                        break
                    if isinstance(n, Loop):
                        break
                    # Don't lift a reservation above a reservation for the same tensor
                    if isinstance(n, TensorHolder) and oset(node.purposes) & oset(
                        n.tensors
                    ):
                        break
                    # Don't lift a reservation above a storage for the same tensor
                    if isinstance(n, Reservation) and oset(n.purposes) & oset(
                        node.purposes
                    ):
                        break
                if not found and node.purposes:
                    new_nodes.append(node)
            else:
                new_nodes.append(node)
            if isinstance(node, MappingNodeWithChildren):
                node._consolidate_reservations()
        assert new_nodes, "BUG"
        self.nodes = EvalableList(new_nodes)

    def _elevate_persistent_nodes_above_splits(self) -> None:
        new_nodes: list[MappingNode] = []
        for node in self.nodes:
            if isinstance(node, Split):
                persistent_nodes = node._get_persistent_nodes()
                new_nodes.extend(persistent_nodes)
                node.clear_nodes(*persistent_nodes)
            if isinstance(node, MappingNodeWithChildren):
                node._elevate_persistent_nodes_above_splits()
            new_nodes.append(node)
        self.nodes = EvalableList(new_nodes)

    def _elevate_tensor_holders_above_splits(self) -> None:
        new_nodes: list[MappingNode] = []
        for node in self.nodes:
            if isinstance(node, Split):
                shared_tensor_holders = node._get_shared_tensor_holders()
                new_nodes.extend(shared_tensor_holders)
                node.clear_nodes(*shared_tensor_holders)
            if isinstance(node, MappingNodeWithChildren):
                node._elevate_tensor_holders_above_splits()
            new_nodes.append(node)
        self.nodes = EvalableList(new_nodes)

    def _propagate_backing_reservations_between_splits(self) -> None:
        for node in self.nodes:
            if isinstance(node, MappingNodeWithChildren):
                node._propagate_backing_reservations_between_splits()

        if not isinstance(self, Split):
            return

        for i, node1 in enumerate(self.nodes):
            # + 2 because if it's in back-to-back branches, then there's nothing in
            # between to add the reservation to.
            for j in range(i + 2, len(self.nodes)):
                node2 = self.nodes[j]
                reservations1 = node1.get_nodes_of_type(Reservation)
                reservations2 = node2.get_nodes_of_type(Reservation)

                reservations1 = [r for r in reservations1 if r._backing]
                reservations2 = [r for r in reservations2 if r._backing]

                shared_reservations = []
                for reservation1 in reservations1:
                    for reservation2 in reservations2:
                        if reservation1 == reservation2:
                            shared_reservations.append(reservation1)
                            break

                for s in shared_reservations:
                    for k in range(i + 1, j):
                        node3 = self.nodes[k]
                        if not isinstance(node3, Nested):
                            raise ValueError(f"Expected Nested node, got {type(node3)}")
                        reservations3 = node3.get_nodes_of_type(Reservation)
                        if s not in reservations3:
                            node3.nodes.insert(0, copy.deepcopy(s))

    def _move_tensor_holders_above_reservations(self) -> None:
        groups = []
        cur_group = []
        for node in self.nodes:
            if isinstance(node, MappingNodeWithChildren):
                node._move_tensor_holders_above_reservations()
            if not isinstance(node, (TensorHolder, Reservation)):
                groups.append(cur_group)
                cur_group = []
            cur_group.append(node)
        groups.append(cur_group)
        groups = [g for g in groups if g]

        groups = [
            [x for x in g if not isinstance(x, (TensorHolder, Reservation))]
            + [x for x in g if isinstance(x, (TensorHolder))]
            + [x for x in g if isinstance(x, (Reservation))]
            for g in groups
        ]
        self.nodes = EvalableList([x for g in groups for x in g])

    def _remove_reservations_for_tolls(self) -> None:
        tolls = self.get_nodes_of_type(Toll)
        toll_names = oset(ps.component for ps in tolls)
        reservations = self.get_nodes_of_type(Reservation)
        remove = [r for r in reservations if r.resource in toll_names]
        self.clear_nodes(*remove)


class Split(MappingNodeWithChildren):
    """
    A :class:`~.MappingNodeWithChildren` that splits the tree into multiple branches,
    each applying to different Einsums.
    """

    def __str__(self) -> str:
        return "Split"

    def _render_node_shape(self) -> str:
        return "hexagon"

    def _get_persistent_nodes(self) -> list[MappingNode]:
        nodes = []
        for n in self.nodes:
            nodes.extend(n.get_nodes_of_type(TensorHolder))
            nodes.extend(n.get_nodes_of_type(Reservation))
        return [n for n in nodes if n.persistent]

    def _get_shared_tensor_holders(self) -> list[TensorHolder]:
        tensor_holders = [n.get_nodes_of_type(TensorHolder) for n in self.nodes]
        shared_tensor_holders = []
        for i in range(len(tensor_holders)):
            for j in range(i + 1, len(tensor_holders)):
                for a in tensor_holders[i]:
                    for b in tensor_holders[j]:
                        if a._backing & b._backing and a not in shared_tensor_holders:
                            assert len(a.tensors) == 1 and len(b.tensors) == 1, "BUG"
                            shared_tensor_holders.append(a)
                            break
        return shared_tensor_holders

    def _render_node_color(self) -> str:
        return "#FFFFE0"


class Nested(MappingNodeWithChildren):
    """
    A :class:`~.MappingNodeWithChildren` that represents a nested set of nodes. Each
    node is the parent of the next node.
    """

    def model_post_init(self, __context__=None) -> None:
        for node in list(self.nodes)[:-1]:
            assert not isinstance(
                node, MappingNodeWithChildren
            ), f"Nested node has a child with children. Only the last child can have children."

    def _parent2child(
        self, parent: MappingNode
    ) -> list[tuple[MappingNode, MappingNode]]:
        parent2child = []
        for node in self.nodes:
            parent2child.append((parent, node))
            parent2child.extend(node._parent2child(parent))
            parent = node._parent2next()
        return parent2child

    def _parent2next(self) -> MappingNode:
        if not self.nodes:
            raise ValueError("Nested node has no children")
        return self.nodes[-1]._parent2next()

    # def _render_connect_children(self, names_lines: list[tuple[str, str]], parent_name: str=None) -> list[str]:
    #     return super()._render_connect_children(names_lines)

    def _render_node_label(self, **kwargs) -> str:
        if not self.nodes:
            raise ValueError("Nested node has no children")
        return self.nodes[0]._render_node_label(**kwargs)

    def _render_node_name(self) -> str:
        if not self.nodes:
            raise ValueError("Nested node has no children")
        return self.nodes[0]._render_node_name()

    def _get_n_shared_loops(self, other: "Nested") -> int:
        my_backing = oset(
            (t, s.component) for s in self._get_backers() for t in s._backing
        )
        other_backing = oset(
            (t, s.component) for s in other._get_backers() for t in s._backing
        )
        shared_backing = my_backing & other_backing

        if not shared_backing:
            return 0

        n_shared_loops = 0
        for i, node in enumerate(self.nodes):
            if isinstance(node, Loop):
                n_shared_loops += 1
            if (
                isinstance(node, Reservation)
                and (node.purpose, node.resource) in shared_backing
            ):
                return n_shared_loops
            if isinstance(node, Split):
                for child in node.nodes:
                    max_child_n_shared_loops = 0
                    try:
                        max_child_n_shared_loops = max(
                            max_child_n_shared_loops, child._get_n_shared_loops(other)
                        )
                    except ValueError:
                        pass
                    return max_child_n_shared_loops + n_shared_loops

        raise ValueError("BUG")

    def _break_into_reorderable_groups(
        self, stop_at_n_loops: int
    ) -> list[list[MappingNode]]:
        # We can reorder loops relative to each other
        groups = []
        cur_group = None

        seen_loops = 0

        if stop_at_n_loops == 0 and not any(
            isinstance(node, Loop) for node in self.nodes
        ):
            return [list(self.nodes)]

        i = 0
        for i, node in enumerate(self.nodes):
            if seen_loops >= stop_at_n_loops:
                break
            is_iteration = isinstance(node, Loop)
            if cur_group is None:
                cur_group = []
            elif (is_iteration and not all(isinstance(x, Loop) for x in cur_group)) or (
                not is_iteration and any(isinstance(x, Loop) for x in cur_group)
            ):
                groups.append(cur_group)
                cur_group = []
            cur_group.append(node)
            assert not isinstance(node, Sequential) or i == len(self.nodes) - 1, "BUG"
            if isinstance(node, Loop):
                seen_loops += 1

        if cur_group:
            groups.append(cur_group)

        final_group = self.nodes[i:]
        groups.append(final_group)

        if seen_loops < stop_at_n_loops:
            raise ValueError(
                f"Expected {stop_at_n_loops} loops, but only found {seen_loops}"
            )

        # Lower reservations. If reservations are in the second-to-last group
        # # non-iteration group, lower them to the last group.
        # if len(groups) > 3:
        #     assert not any(isinstance(x, Loop) for x in groups[-1]), "BUG"
        #     assert not any(isinstance(x, Loop) for x in groups[-3]), "BUG"
        #     reservations = [x for x in groups[-2] if isinstance(x, Reservation)]
        #     groups[-1].extend(reservations)
        #     groups[-3] = [x for x in groups[-3] if x not in reservations]

        return groups

    def _merge(self, other: "Nested", n_shared_loops: int) -> "Nested":

        # Break up the nodes above the indices. We need to have them in the format of
        # [(loop, other stuff...), (loop, other stuff...), ...]
        my_groups = self._break_into_reorderable_groups(stop_at_n_loops=n_shared_loops)
        my_remaining = my_groups.pop(-1)
        other_groups = other._break_into_reorderable_groups(
            stop_at_n_loops=n_shared_loops
        )
        other_remaining = other_groups.pop(-1)

        # Reorder so that the loops are in the same order. We can't reorder groups that
        # have other stuff in them because that'll change the behavior of the mapping.
        zipped_groups = []

        def _pop_loop_group(groups: list[list[MappingNode]]) -> list[MappingNode]:
            while groups and not any(isinstance(x, Loop) for x in groups[0]):
                zipped_groups.append(groups.pop(0))
            return groups.pop(0) if groups else []

        my_loop_group = _pop_loop_group(my_groups)
        other_loop_group = _pop_loop_group(other_groups)
        while (my_groups or my_loop_group) and (other_groups or other_loop_group):
            if not my_loop_group:
                my_loop_group = _pop_loop_group(my_groups)
                continue
            if not other_loop_group:
                other_loop_group = _pop_loop_group(other_groups)
                continue

            # Add matching loops from the two groups. If we can't find a match, raise an
            # error.
            to_add = None
            for i, a in enumerate(my_loop_group):
                for j, b in enumerate(other_loop_group):
                    if a == b:
                        to_add = [a]
                        my_loop_group.pop(i)
                        other_loop_group.pop(j)
                        break
                if to_add is not None:
                    break

            if to_add is None:
                # TODO: This check for one is only to early catch bugs coming here. The
                # code below says that if we couldn't find a match, then ignore rank
                # variables and assume that rank variable translation would fix it.
                assert len(my_loop_group) == 1 or len(other_loop_group) == 1
                has_one, may_not_have_one = my_loop_group, other_loop_group
                has_one_side, may_not_side = self, other
                if len(has_one) != 1:
                    has_one, may_not_have_one = may_not_have_one, has_one
                    has_one_side, may_not_side = may_not_side, has_one_side

                l = copy.deepcopy(has_one.pop(0))
                l.rank_variable = _as_oset(l.rank_variable)
                for l2 in may_not_have_one:
                    if l2.calculated_n_iterations == l.calculated_n_iterations:
                        break
                else:
                    raise ValueError(
                        f"No matching loop found for {my_loop_group} and {other_loop_group}"
                    )
                print(
                    f"Warning. Matching loops {l} and {l2}. Need rank variable translation here."
                )

                # For each einsum, record which rank variable it used for this
                # loop.
                for src_loop, src_side in [(l, has_one_side), (l2, may_not_side)]:
                    src_rv = _as_oset(src_loop.rank_variable)
                    for compute in src_side.get_nodes_of_type(Compute):
                        rv = src_loop._einsum_to_rank_variable.get(compute.einsum)
                        if rv is None:
                            assert len(src_rv) == 1, (
                                f"Loop {src_loop} has no mapping for einsum "
                                f"{compute.einsum} and has multiple rank "
                                f"variables {src_rv}. A prior merge should have "
                                f"populated _einsum_to_rank_variable."
                            )
                            rv = next(iter(src_rv))
                        l._einsum_to_rank_variable[compute.einsum] = rv

                may_not_have_one.remove(l2)
                l.rank_variable = l.rank_variable | _as_oset(l2.rank_variable)
                to_add = [l]

            zipped_groups.append(to_add)

        assert not my_loop_group and not other_loop_group, "BUG"

        zipped_groups.extend(my_groups)
        zipped_groups.extend(other_groups)

        flattened = list(x for group in zipped_groups for x in group)
        new_nodes = [x for x in flattened if not isinstance(x, Sequential)]
        new_nodes.extend([x for x in flattened if isinstance(x, Sequential)])

        if isinstance(my_remaining[0], Sequential) and isinstance(
            other_remaining[0], Sequential
        ):
            my_remaining[0].nodes.extend(other_remaining[0].nodes)
            assert len(my_remaining) == 1 and len(other_remaining) == 1, "BUG"
            new_nodes.append(my_remaining[0])
        elif isinstance(my_remaining[0], Sequential):
            my_remaining[0].nodes.append(Nested(nodes=other_remaining))
            assert len(my_remaining) == 1, "BUG"
            new_nodes.append(my_remaining[0])
        elif isinstance(other_remaining[0], Sequential):
            other_remaining[0].nodes.append(Nested(nodes=my_remaining))
            assert len(other_remaining) == 1, "BUG"
            new_nodes.append(other_remaining[0])
        else:
            new_nodes.append(
                Sequential(
                    nodes=[Nested(nodes=my_remaining), Nested(nodes=other_remaining)]
                )
            )

        return Nested(nodes=new_nodes)

    def _beautify_loops(
        self, rank_variable_bounds: Optional[dict[str, dict[str, int]]] = None
    ):
        to_remove = []
        rank_variable_bounds = rank_variable_bounds or {}

        for i, node in enumerate(self.nodes):
            if not isinstance(node, Loop):
                continue
            prev_tile_shape = None
            for j in range(i - 1, -1, -1):
                node2 = self.nodes[j]
                if not isinstance(node2, Loop):
                    continue
                if node2.tile_shape is None:
                    continue
                if node2.rank_variable != node.rank_variable:
                    continue
                prev_tile_shape = node2.tile_shape
                break
            if prev_tile_shape is None:
                prev_tile_shape = rank_variable_bounds.get(node.rank_variable, None)
            if prev_tile_shape is not None:
                if node.tile_shape == prev_tile_shape:
                    to_remove.append(i)
                    continue
                elif node.tile_shape is not None and prev_tile_shape is not None:
                    node.tile_pattern = node.tile_pattern.update(
                        calculated_n_iterations=ceil(prev_tile_shape / node.tile_shape),
                    )

        def safe_int_cast(x: int | float | None) -> int | float | None:
            try:
                int_x = int(x)
                return int_x if int_x == x else x
            except:
                pass
            return x

        for i, node in enumerate(self.nodes):
            if not isinstance(node, Loop):
                continue
            node.tile_pattern = node.tile_pattern.update(
                initial_tile_shape=safe_int_cast(node.tile_pattern.initial_tile_shape),
                tile_shape=safe_int_cast(node.tile_pattern.tile_shape),
            )

        self.nodes = [node for i, node in enumerate(self.nodes) if i not in to_remove]

    @override
    def compact_str(self) -> str:
        result = []
        prev = None
        for node in self.nodes:
            try:
                prev = prev._merge(node)
            except:
                if prev is not None:
                    result.append(prev)
                prev = node
        if prev is not None:
            result.append(prev)

        return " ".join(f"{node.compact_str()}" for node in result)

    def _get_single_tensor_mapping(
        self,
        tensor_name: TensorName,
        flattened_arch: list[arch.Leaf],
        tensor_rank_variables: set[str],
    ) -> Self:
        """
        Ctrl-F for CONTIGUOUS_ITERATION_SPACE_DISCUSSION

        Returns this Nested node with only the nodes associated with the given tensor.

        Includes loops and compute nodes, plus any tensor holders and reservations that
        are associated with the given tensor.

        Puts spatials as high as they can go while being below any node that is above
        them in the memory hierarchy. Between two tensor holders, generally puts spatial
        loops at the bottom, but may put them above temporal loops if that better lines
        up with the original order. When memory hierarchy order is followed globally
        (e.g., output in buffer must be above input in reg), the loop order going into
        this function will always match that going out.

        This function expects, as input, all spatials to be placed as low as they can
        go, but above their respective fanouts.

        When memory hierarchy order is only followed per-tensor (e.g., output in buffer
        must be above output in reg, but can be below input in reg), things may be more
        complicated. We discuss this in more detail using the following example:

        Hierarchy:

        - Buffer
        - 2x fanout
        - Reg

        Mapping:
            S-reg for m1 in [0, 2):
              [Reg reuses input]
                for m0 in [0, 2):
                  [Buffer reuses output]

        When given a mapping and architecture like the above, this function may reorder
        the spatial and temporal loops, yielding the following:

        Mapping for input:
            S-reg for m1 in [0, 2):
              [Reg reuses input]
                for m0 in [0, 2):

        Mapping for output:
            for m0 in [0, 2):
              [Buffer reuses output]
                S-reg for m1 in [0, 2):

        Unfortunately, such reordering is inevitable given our assumptions of an
        inclusive memory hierarchy (because any tile stored in the reg must be stored in
        the buffer), and our desire to place the reg storage node higher. It's also a
        symptom of the following other issues:

        - In cases like these, storage nodes may need to keep non-contiguous chunks of
          the iteration space. For example, if the spatial loop is on top, then one reg
          holds [0, 1] while the other holds [2, 3]. Meanwhile, in the first temporal
          iteration, the buffer holds [0, 2] and in the second temporal iteration, the
          buffer holds [2, 4].
        - We get weird dependencies between loop order and compatibility for fusion
          because loop order affects the iteration space tiles that are stored.

        To prevent these problems from occuring, we raise an error if there any temporal
        loops in between that affect the same indexing expressions as the spatial loops.
        I tried to have it work with our model and then constraining the temporal loops
        to be null (have the same tile shape as their outer loop), but when we run it
        per-tensor and reorder, the loop above the temporal changes, so the model
        returns inconsistent results for each tensor as the tile shape is different.
        With this constraint, we'll never reorder spatial and temporal loops that affect
        one another.

        The result of the above is that we'll never reorder spatial and temporal loops
        that affect one another.

        I haven't thought through how this will work with more complex rank variable
        expressions, so to be safe, will say that there can not be a temporal and
        spatial loop that affect the same indexing expression or each others' loop
        bounds.

        These problems also aren't necessarily impossible to solve; I just haven't
        thought it through. If we do think it through, a good place to start would be to
        update the model to support non-contiguous chunks of the iteration space, then
        come up with some way to explore mappings and fusion while using non-contiguous
        chunks of the iteration space.

        TODO: Mapper then also needs explore swapping temporal/spatial loops
        """
        spatials = [n for n in self.nodes if isinstance(n, Spatial)]
        tensor_holders = [
            n for n in self.nodes if isinstance(n, (TensorHolder, Compute))
        ]
        others = [
            n
            for n in self.nodes
            if not isinstance(n, (TensorHolder, Reservation, Spatial))
            or (isinstance(n, TensorHolder) and n.tensor == tensor_name)
            or (isinstance(n, Reservation) and n.purpose == tensor_name)
        ]
        assert not any(isinstance(n, MappingNodeWithChildren) for n in others), "BUG"

        def arch_idx(node: MappingNode) -> int:
            for i, n in enumerate(flattened_arch):
                if n.name == node.component:
                    return i
            raise ValueError(f"Component {node.component} not found in flattened arch")

        spatials_above = {
            id(node): [s for s in spatials if arch_idx(s) <= arch_idx(node)]
            for node in tensor_holders
        }
        spatials_below = {
            id(node): [s for s in spatials if arch_idx(s) >= arch_idx(node)]
            for node in tensor_holders
        }

        mapping = []
        for to_add in others:
            if isinstance(to_add, (TensorHolder, Compute)):
                cur_spatials_above = [
                    s for s in spatials if s in spatials_above[id(to_add)]
                ]
                spatials = [s for s in spatials if s not in cur_spatials_above]
                mapping.extend(cur_spatials_above)
            mapping.append(to_add)

        mapping.extend(spatials)

        # Check that spatials are always above their respective fanouts
        for i, node in enumerate(mapping):
            if not isinstance(node, (TensorHolder, Compute)):
                continue
            for node2 in mapping[i + 1 :]:
                if not isinstance(node2, Spatial):
                    continue
                assert node2 in spatials_below[id(node)], "BUG"
                assert node2 not in spatials_above[id(node)], "BUG"

        # Split the mapping into groups of tensor holders and sequential loops
        id2idx = {id(node): i for i, node in enumerate(self.nodes)}
        groups = []
        for node in mapping:
            if (
                isinstance(node, Loop)
                and len(groups) > 0
                and isinstance(groups[-1][0], Loop)
            ):
                groups[-1].append(node)
            else:
                groups.append([node])

        groups = [sorted(g, key=lambda x: id2idx[id(x)]) for g in groups]
        mapping = [x for g in groups for x in g]

        # Check that all storage-temporal relations are held from before.
        # Precomputed id-sets avoid pydantic __instancecheck__ in the O(N²)
        # combinations loops.
        node2idx = {id(node): i for i, node in enumerate(mapping)}
        prev_node2idx = {id(node): i for i, node in enumerate(self.nodes)}
        th_ids = {id(n) for n in mapping if isinstance(n, TensorHolder)}
        sp_ids = {id(n) for n in mapping if isinstance(n, Spatial)}
        loop_ids = {id(n) for n in mapping if isinstance(n, Loop)}
        for n1, n2 in itertools.combinations(mapping, 2):
            i1, i2 = id(n1), id(n2)
            if (i1 in th_ids and i2 in th_ids) or (i1 in sp_ids and i2 in sp_ids):
                assert (node2idx[i1] > node2idx[i2]) == (
                    prev_node2idx[i1] > prev_node2idx[i2]
                ), "BUG"

        # Check for spatial/temporal loops that have been reordered for the
        # same rank variable. These can't co-exist because the tiling is
        # inconsistent. OK for irrelevant loops.
        node2idx = {id(node): i for i, node in enumerate(self.nodes)}
        for node1, node2 in itertools.combinations(mapping, 2):
            if id(node1) not in loop_ids or id(node2) not in loop_ids:
                continue
            if node2idx[id(node1)] <= node2idx[id(node2)]:
                continue
            if node1.rank_variable not in tensor_rank_variables:
                continue
            if node1.rank_variable != node2.rank_variable:
                continue
            s = """
            In the given mapping, there exists (potentially with other nodes in
            between) a spatial loop above a temporal loop above a storage node,
            where the loops index into the same indexing expression, and the storage
            node is not fanned out by the spatial loop. This is not allowed.

            Mapping:
            """
            s = s.replace("                ", "")

            to_add = []
            for n in self.nodes:
                if id(n) == id(node1) or id(n) == id(node2):
                    to_add.append(f"\t{n.compact_str()} <-- Offending Loop")
                else:
                    to_add.append(f"\t{n.compact_str()}")

            raise ValueError(s + "\n".join(to_add))

        return type(self)(nodes=mapping)


class _Parallel(Split):
    """
    A :class:`~.Split` where each branch operates at the same time in different
    spatially-organized hardware.
    """

    pass


class Pipeline(Split):
    """
    A :class:`~.Split` where each branch operates at the same time in different
    spatially-organized hardware.
    """

    pass


class Sequential(Split):
    """
    A :class:`~.Split` where branches are processed one-after-another.
    """

    pass


# =============================================================================
# Nodes That May Only be Inserted by the Model
# =============================================================================


class Reservation(MappingNode):
    """A node that reserves a hardware resource for a specific task."""

    purposes: EvalableList[str]
    """ The reasons for reserving the resource. """

    resource: str
    """ The resource being reserved. """

    _backing: Set[str] = oset()
    """ Tensors for which this reservation is reserving the tensor's backing storage.
    """

    persistent: bool = False
    """
    Whether this reservation is persistent. Persistent reservations can't be tiled and
    must be kept in backing storage for the full duration of the workload's execution.
    """

    @override
    def compact_str(self) -> str:
        return f'{",".join(self.purposes)} reserves {self.resource}'

    def __str__(self, color_map: ColorMap = None) -> str:
        purposes = self.purposes
        if color_map is not None:
            format_list = [f"{self.resource} reserved for"] + list(purposes)
            return color_map.format_list(format_list)
        return f"{self.resource} reserved for {",".join(purposes)}"

    def _render_node_shape(self) -> str:
        return "component"

    @property
    def purpose(self) -> str:
        if len(self.purposes) == 1:
            return self.purposes[0]
        raise ValueError(f"Reservation has multiple purposes: {self.purposes}")

    def __eq__(self, other: "Reservation") -> bool:
        return (
            isinstance(other, Reservation)
            and self.purposes == other.purposes
            and self.resource == other.resource
        )

    def _render_node_color(self) -> str:
        return "#E8E8E8"  # Light gray


# =============================================================================
# Top-level Mapping
# =============================================================================

MappingNodeTypes: TypeAlias = Union[
    Temporal,
    Spatial,
    Storage,
    Pipeline,
    Sequential,
    Compute,
    Reservation,
    # Fill,
    TensorHolder,
]
"""TypeAlias MappingNodeTypes: The types of MappingNodes possible."""


class Mapping(Nested):
    """A Mapping of a workload onto a hardware architecture."""

    _n_loop_orders: int | None = None
    """ Used for counting number of unique mappings. Do not touch. """

    _template_index: int | None = None
    """ Used for tracking which mapping is which. """

    def remove_reservations(self):
        self.nodes = [n for n in self.nodes if not isinstance(n, Reservation)]

    def split_reservations(self):
        new_nodes = []
        for node in self.nodes:
            if isinstance(node, Reservation):
                for purpose in node.purposes:
                    new_node = copy.copy(node)
                    new_node.purposes = [purpose]
                    new_nodes.append(new_node)
            else:
                new_nodes.append(node)
        self.nodes = new_nodes

    def split_loop_with_multiple_rank_variables(self, einsum_name: EinsumName):
        new_nodes = []
        for node in self.nodes:
            if isinstance(node, Loop) and isinstance(node.rank_variable, set):
                if einsum_name not in node._einsum_to_rank_variable:
                    raise ValueError(
                        f"Loop {node.compact_str()} has multiple rank variables. Make "
                        f"sure to populate _einsum_to_rank_variable for this loop."
                    )
                new_node = copy.copy(node)
                new_node.rank_variable = node._einsum_to_rank_variable[einsum_name]
                new_nodes.append(new_node)
            else:
                new_nodes.append(node)
        self.nodes = new_nodes

    def clear_irrelevant_reservations(self, relevant_tensors: set[TensorName]):
        new_nodes = []
        for node in self.nodes:
            if isinstance(node, Reservation):
                if node.purpose not in relevant_tensors:
                    continue
                new_nodes.append(node)
            elif isinstance(node, TensorHolder):
                node.tensors = [t for t in node.tensors if t in relevant_tensors]
                if node.tensors:
                    new_nodes.append(node)
            elif isinstance(node, MappingNodeWithChildren):
                node.clear_irrelevant_reservations(relevant_tensors)
                new_nodes.append(node)
            else:
                new_nodes.append(node)
        self.nodes = new_nodes

    def split_tensor_holders_with_multiple_tensors(self):
        new_nodes = []
        for node in self.nodes:
            if isinstance(node, TensorHolder) and len(node.tensors) > 1:
                for tensor in node.tensors:
                    new_node = copy.copy(node)
                    new_node.tensors = [tensor]
                    new_nodes.append(new_node)
            else:
                new_nodes.append(node)
        self.nodes = new_nodes

    def _get_fused_slice(self, fusable_tensors: set[TensorName]) -> "Mapping":
        """
        Return a mapping with:
        - All backing reservation nodes for intermediate tensors
        - Loop nodes above any backing reservation nodes
        """
        # All intermediate tensors that can be found in this mapping
        # Note: `fusable_tensors` may be for **whole workload**.
        relevant_intermediate_tensors = oset()
        for node in self.nodes:
            if isinstance(node, Reservation):
                if node.purpose in fusable_tensors:
                    relevant_intermediate_tensors.add(node.purpose)

        fused_slice = Mapping(nodes=[])
        to_add = []
        for node in self.nodes:
            node = copy.copy(node)
            if isinstance(node, Reservation):
                if node.purpose not in relevant_intermediate_tensors:
                    continue
                fused_slice.nodes.extend(to_add + [node])
                to_add = []
                relevant_intermediate_tensors.remove(node.purpose)
                if len(relevant_intermediate_tensors) == 0:
                    break
            elif isinstance(node, Loop):
                to_add.append(node)
        return fused_slice

    @property
    def loops(self) -> list[Loop]:
        """Returns all :class:`~.Loop` nodes in the Mapping."""
        return self.get_nodes_of_type(Loop)

    def _render_node_label(self, **kwargs) -> str:
        return f"Root"

    def _repr_svg_(self) -> str:
        return self.render()

    def render_pydot(self, with_reservations=True, with_tile_shape=True) -> pydot.Dot:
        """Renders the mapping as a Pydot graph. Returns an SVG string."""
        graph = _pydot_graph()
        # Enable HTML-like labels for color support
        graph.set_node_defaults(label="")
        if not with_reservations:
            exclude_types = (Reservation,)
        else:
            exclude_types = tuple()
        for node in self._render_make_children(
            exclude_types=exclude_types, with_tile_shape=with_tile_shape
        ):
            graph.add_node(node)

        color_keys = oset()
        all_nodes = self._flatten()
        for node in all_nodes:
            if isinstance(node, TensorHolder):
                color_keys.update(node.tensors)
            if isinstance(node, Reservation):
                color_keys.update(node.purposes)

        color_map = ColorMap(sorted(color_keys))

        for node in all_nodes:
            if isinstance(node, (TensorHolder, Reservation)):
                graph_nodes = graph.get_node(node._render_node_name())
                for graph_node in graph_nodes:
                    # Set HTML-like label for color support
                    new_label = node.__str__(color_map)
                    graph_node.set_label(new_label)
                    # graph_node.set_fillcolor(color_map[node._color_key()])
                    # graph_node.set_style('filled')

        added_edges = oset()
        child2included_parent = {}
        for parent, child in self._parent2child(None):
            parent_name = parent._render_node_name() if parent is not None else None
            child_name = child._render_node_name()
            if isinstance(parent, exclude_types):
                parent_name = child2included_parent.get(parent_name, None)
            child2included_parent[child_name] = parent_name
            if not isinstance(child, exclude_types):
                added_edges.add((parent_name, child_name))
        for parent_name, child_name in added_edges:
            if parent_name is not None:
                graph.add_edge(pydot.Edge(parent_name, child_name))
        return graph

    def render(self, with_reservations=True, with_tile_shape=True) -> _SVGJupyterRender:
        graph = self.render_pydot(with_reservations, with_tile_shape)
        return _SVGJupyterRender(graph.create_svg(prog="dot").decode("utf-8"))

    @classmethod
    def _from_pmappings(
        cls,
        pmappings: list[Nested],
        rank_variable_bounds: Optional[dict[str, dict[str, int]]] = None,
        _for_model: bool = False,
    ) -> "Mapping":
        orig_pmappings = pmappings
        pmappings = list(copy.deepcopy(pmappings))
        for pmapping in pmappings:
            pmapping._beautify_loops(rank_variable_bounds)

        import os

        no_join = (
            _NO_JOIN_MAPPING_VISUALIZATION
            or os.environ.get("NO_JOIN_MAPPING_VISUALIZATION", "False") == "True"
        )

        while len(pmappings) > 1:
            highest_n_shared_loops = 0
            highest_shared_pmapping_index = 0
            for i, pmapping in enumerate(pmappings):
                shared_index = 0
                for j in range(i + 1, len(pmappings)):
                    shared_index = max(
                        pmapping._get_n_shared_loops(pmappings[j]), shared_index
                    )
                if shared_index > highest_n_shared_loops:
                    highest_n_shared_loops = shared_index
                    highest_shared_pmapping_index = i

            def einsum_names(pmapping: Nested) -> str:
                return ",".join(n.einsum for n in pmapping.get_nodes_of_type(Compute))

            names_a = einsum_names(pmappings[highest_shared_pmapping_index])
            names_b = einsum_names(pmappings[highest_shared_pmapping_index + 1])
            # print(
            #     f"Merging with shared loops {highest_n_shared_loops}: {names_a} <--> {names_b}."
            # )
            # print(pmappings[highest_shared_pmapping_index]._get_n_shared_loops(pmappings[highest_shared_pmapping_index + 1]))
            pmappings[highest_shared_pmapping_index] = pmappings[
                highest_shared_pmapping_index
            ]._merge(
                pmappings.pop(highest_shared_pmapping_index + 1),
                0 if no_join else highest_n_shared_loops,
            )

        mapping: Mapping = cls(nodes=pmappings[0].nodes)
        mapping._elevate_persistent_nodes_above_splits()

        # This was messing up the model because it could do relative reordering of
        # storage nodes. Specifically, in copy ops, it would cause problems in the
        # following case:
        #
        # [Copy source] [Copy destination]
        #
        # Where copy destination is shared between multiple einsums. Then, source and
        # destination would both be below the split, causing destination to be elevated
        # above, and the actual copy operation to move things from destination ->
        # source.
        if not _for_model:
            mapping._elevate_tensor_holders_above_splits()
        mapping._propagate_backing_reservations_between_splits()
        mapping._consolidate_tensor_holders()
        mapping._consolidate_reservations()
        mapping._move_tensor_holders_above_reservations()
        # mapping._remove_reservations_for_tolls()
        return mapping

        # import mermaid as md
        # from mermaid.graph import Graph
        # lines = []
        # lines = [
        #     "graph TD",
        #     "%%{init: {'flowchart': {'nodeSpacing': 30, 'rankSpacing': 30, 'padding': 2}, 'themeVariables': {'fontFamily': 'Arial, sans-serif'}}}%%"
        # ]
        # lines.extend(self._render_make_children())
        # for parent, child in self._parent2child(None):
        #     if parent is not None:
        #         lines.append(f"{parent._render_node_name()} --> {child._render_node_name()}")
        #     # if _is_root:
        # #     lines.extend([
        # #         "",
        # #         "classDef default fill:#fff,stroke:#000,stroke-width:1px,color:#000,font-family:Arial,font-size:12px,padding:2px;",
        # #         "classDef compact fill:#fff,stroke:#000,stroke-width:1px,color:#000,font-family:Arial,font-size:12px,padding:2px;"
        # #     ])

        # # Create the graph with the flowchart script
        # flowchart_script = "\n".join(lines)
        # graph = Graph('Flowchart', flowchart_script)

        # # Set the configuration for compact layout
        # config = md.Config()
        # config.theme = 'base'
        # # config.theme_variables = {
        # #     'primaryColor': '#ffffff',
        # #     'primaryTextColor': '#000000',
        # #     'primaryBorderColor': '#000000',
        # #     'lineColor': '#000000',
        # #     'fontSize': '12px'
        # # }
        # # config.flowchart = {
        # #     'nodeSpacing': 20,
        # #     'rankSpacing': 10,
        # #     'curve': 'linear'
        # # }
        # graph.config = config

        # return md.Mermaid(graph)


Split.model_rebuild()
Nested.model_rebuild()
Mapping.model_rebuild()



"""
ISL functions that encapsulate more commonly used workflows in looptree for the
sake of code concision.
"""

import logging
from typing import List

import islpy as isl


def project_dim_in_after(map_: isl.Map, start: int) -> isl.Map:
    """
    Projects out the input dims of idx [start, end] in map_.

    Parameters
    ----------
    map_:
        The map to project out dims from.
    start:
        The dim idx to start projecting dims out from.

    Returns
    -------
    map_ without the input dims [start:].
    """
    n_dim_in: int = map_.dim(isl.dim_type.in_)
    return (
        map_.project_out(isl.dim_type.in_, start, n_dim_in - start)
        if start <= n_dim_in
        else map_
    )


def dim_projector_range(space: isl.Space, start: int, n: int) -> isl.Map:
    """
    Given a space, create a map that projects out the dims [start: start+n).

    Parameters
    ----------
    space:
        The space to create the dim projector in.
    start:
        The index to start the projection.
    n:
        The number of dims from `start` to project out.

    Returns
    -------
    A `isl.Map` in `space` that projects out dims [start:start+n].
    """
    base_map: isl.Map = isl.Map.identity(isl.Space.map_from_set(space))
    # TODO: propagate tuple names from `space` onto `base_map` (e.g.,
    # `space.get_tuple_name(isl.dim_type.set)`) so the projector keeps the
    #  original set label on both domain and range.
    return isl.Map.project_out(base_map, isl.dim_type.in_, start, n)


def dim_projector_mask(space: isl.Space, mask: List[bool]) -> isl.Map:
    """
    Given a space, create a map that projects out the dims marked `True` in the mask.

    Parameters
    ----------
    space:
        The space the projector is created on.
    mask:
        The mask of the list of dims to be projected out.

    Returns
    -------
    A projection from space in[x_0, ..., x_n] -> out[x_1, ..., x_i, ... x_{n-1}]
    where `x_i ∉ out => mask[i]`.
    """
    projector: isl.Map = isl.Map.identity(isl.Space.map_from_set(space))
    # TODO: set tuple names on `projector` using the set name (e.g.,
    #  `space.get_tuple_name(isl.dim_type.set)`) so domain/range remain
    # attributable after masking.

    for i in range(len(mask) - 1, -1, -1):
        if mask[i]:
            projector = projector.project_out(isl.dim_type.in_, i, 1)

    return projector


def add_dims_preserve_name_map(
    map_: isl.Map, dim_type: isl.dim_type, n: int
) -> isl.Map:
    """
    Wrapper of `isl.Map.add_dims` that preserves the space name post
    addition.

    Parameters
    ----------
    map_:
        The map we're adding into.
    dim_type:
        The dimension tuple we're inserting into.
    n:
        The number of dimensions to insert.

    Returns
    -------
    Dimension-inserted maps with preservation.

    Postcondition
    -------------
    The entirety of dependencies with a given space name in the given context
    we're operating under.
    """
    name: str = map_.get_tuple_name(dim_type)
    map_ = map_.add_dims(dim_type, n)
    if name is None:
        logging.warning(f"unnamed space for {map_}", stack_info=True)
        return map_
    return map_.set_tuple_name(dim_type, name)


def insert_dims_preserve_name_map(
    map_: isl.Map, dim_type: isl.dim_type, pos: int, n: int
) -> isl.Map:
    """
    Wrapper of `isl.Map.insert_dims` that preserves the space name post
    insertion.

    Parameters
    ----------
    map_:
        The inserting map.
    dim_type:
        The dimension tuple we're inserting into.
    pos:
        The position to start inserting into, inclusive.
    n:
        The number of dimensions to insert.

    Returns
    -------
    Dimension-inserted maps with preservation.

    Postcondition
    -------------
    The entirety of dependencies with a given space name in the given context
    we're operating under.
    """
    name: str = map_.get_tuple_name(dim_type)
    map_ = map_.insert_dims(dim_type, pos, n)
    if name is None:
        logging.warning(f"unnamed space for {map_}", stack_info=True)
        return map_
    return map_.set_tuple_name(dim_type, name)


def insert_equal_dims_maff(
    maff: isl.MultiAff, in_pos: int, out_pos: int, n: int
) -> isl.MultiAff:
    """
    Given a multi affine, insert equal numbers of input and output dimensions and
    enforce equality between the values of the two dims.

    Parameters
    ----------
    maff:
        The multi affine base to insert dims into.
    in_pos:
        The index to start inserting input dimensions at in `maff`.
    out_pos:
        The index to start inserting output dimensions at in `maff`.
    n:
        The number of dimensions to insert.

    Returns
    -------
    A new maff which is equivalent to `maff` except it has `n` new input and
    output dimensions starting at `in_pos` and `out_pos` respectively.
    """
    # Inserts the `n` dimensions into a new maff base.
    maff = maff.insert_dims(isl.dim_type.in_, in_pos, n)
    maff = maff.insert_dims(isl.dim_type.out, out_pos, n)

    # Modifies each affine to create an equality relation between the input and output.
    for i in range(n):
        aff: isl.Aff = maff.get_at(out_pos + i)
        aff = aff.set_coefficient_val(isl.dim_type.in_, in_pos + i, 1)
        maff = maff.set_aff(out_pos + i, aff)

    return maff


def insert_equal_dims_map(map_: isl.Map, in_pos: int, out_pos: int, n: int) -> isl.Map:
    """
    Given a map, insert equal numbers of input and output dimensions and enforce
    equality between the values of the two dims.

    Parameters
    ----------
    map_:
        The map base to insert dims into.
    in_pos:
        The index to start inserting input dimensions at in `map_`.
    out_pos:
        The index to start inserting output dimensions at in `map_`.
    n:
        The number of dimensions to insert.

    Returns
    -------
    A new maff which is equivalent to `map_` except it has `n` new input and
    output dimensions starting at `in_pos` and `out_pos` respectively.
    """
    # Inserts the new input and output dimensions.
    map_ = insert_dims_preserve_name_map(map_, isl.dim_type.in_, in_pos, n)
    map_ = insert_dims_preserve_name_map(map_, isl.dim_type.out, out_pos, n)

    # Adds constraints for conservation between the new input and output dimensions
    # in the map.
    local_space: isl.LocalSpace = map_.get_space().to_local_space()
    for i in range(n):
        # out - in == 0 => out == in
        constraint: isl.Constraint = isl.Constraint.alloc_equality(local_space)
        constraint = constraint.set_coefficient_val(isl.dim_type.in_, in_pos + i, 1)
        constraint = constraint.set_coefficient_val(isl.dim_type.out, out_pos + i, -1)
        map_ = map_.add_constraint(constraint)

    return map_


def map_to_prior_coordinate(n_in_dims: int, shifted_idx: int, name: str) -> isl.Map:
    """
    Create a map that relates current time index vector to a previous index vector.
    It shifts the coordinate at shifted_idx back by 1.

    Goal: { [i0,...,i{n_in_dims-1}] ->
            [i0, ..., i{shifted_idx}-1, i{shifted_idx+1}, ..., i{n_in_dims-1}] }

    Parameters
    ----------
    n_in_dims:
        The number of input/output dims of the dataspace.
    shifted_idx:
        The coordinate being shifted.
    name:
        The name for the domain and range of the shifter.

    Returns
    -------
    A map relating a current index vector to a previous index to a previous
    index vector by shifting the coordinate at `shifted_idx` back by 1.


    Preconditions
    -------------
    -   0 <= shifted_idx <=n_in_dims-1
    """

    # Creates the space, map, and local_space the temporal reuse data map will exist
    # in.
    space: isl.Space = isl.Space.alloc(isl.DEFAULT_CONTEXT, 0, n_in_dims, n_in_dims)
    map_: isl.Map = isl.Map.empty(space)
    local_space: isl.LocalSpace = isl.LocalSpace.from_space(space)

    constraint: isl.Constraint
    # If there is any data replacement
    if shifted_idx > 0:
        # Create a temporary map.
        tmp_map: isl.Map = isl.Map.universe(space)
        # Model the conservation of data along each data dimension in that map.
        # out - in == 0 => out == in
        for i in range(shifted_idx - 1):
            constraint = isl.Constraint.alloc_equality(local_space)
            constraint = constraint.set_coefficient_val(isl.dim_type.out, i, 1)
            constraint = constraint.set_coefficient_val(isl.dim_type.in_, i, -1)
            tmp_map = tmp_map.add_constraint(constraint)

        # Sets constraints such that the pivot value is decremented.
        # out - in + 1 == 0 => out == in - 1
        constraint = isl.Constraint.alloc_equality(local_space)
        constraint = constraint.set_coefficient_val(
            isl.dim_type.out, shifted_idx - 1, 1
        )
        constraint = constraint.set_coefficient_val(
            isl.dim_type.in_, shifted_idx - 1, -1
        )
        constraint = constraint.set_constant_val(1)
        tmp_map = tmp_map.add_constraint(constraint)

        map_ = map_.union(tmp_map)

    # If we're pivoting any of the data, preserve the `shifted_idx` datapoints.
    if shifted_idx < n_in_dims:
        tmp_map: isl.Map = isl.Map.lex_gt(
            isl.Space.set_alloc(
                isl.DEFAULT_CONTEXT,
                map_.dim(isl.dim_type.param),
                n_in_dims - shifted_idx,
            )
        )
        tmp_map = insert_equal_dims_map(tmp_map, 0, 0, shifted_idx)
        map_ = map_.union(tmp_map)

    map_ = map_.set_tuple_name(isl.dim_type.in_, name).set_tuple_name(
        isl.dim_type.out, name
    )

    return map_


def map_to_shifted(domain_space: isl.Space, pos: int, shift: int) -> isl.Map:
    """
    Given a `domain_space`, return a map from a point in the `domain_space` to
    a point in that dimension `shift` ahead in the dimension at `pos`.

    Parameters
    ----------
    domain_space:
        The space on which to construct the shift on.
    pos:
        The dimension to construct the shift on.
    shift:
        The amount of shift

    Returns
    -------
    A mapping from `[x_0, x_1, ..., x_n] -> [x_0, ..., x_{pos} + shift, ..., x_n]`
    on `domain_space`.
    """
    maff: isl.MultiAff = domain_space.identity_multi_aff_on_domain()
    maff = maff.set_at(pos, maff.get_at(pos).set_constant_val(shift))
    return maff.as_map()


def reorder_projector(
    permutation: list[int], space: str, ctx: isl.Context = isl.DEFAULT_CONTEXT
) -> isl.Map:
    """
    A projection to reorder a space from [i_2, i_0, ..., i_n, ...] -> [i_0, i_1, ..., i_n]

    Parameters
    ----------
    permutation:
        A list where the elements correspond to indices of the space's dimensions
        and the ordering of the indices the reconstruction.
    space:
        The name of the space being permuted.

    Returns
    -------
    A map that reorders an arbitrary list of input dimensions into an enforced ordering.
    """
    # Constructs the permutation in the form of a string, because it's easier
    # and less operations than explicit construction via objects.
    pattern: str
    if len(permutation) == 0:
        pattern = "{ [] -> [] }"
    else:
        pattern = "{ [ "
        for i in range(len(permutation) - 1):
            dim_idx = permutation[i]
            pattern += f"i{dim_idx}, "
        pattern += f"i{permutation[-1]}] -> [ "

        for i in range(len(permutation) - 1):
            pattern += f"i{i}, "

        pattern += f"i{permutation[-1]} ] }}"

    # Creates the actual map.
    reorder_proj: isl.Map = isl.Map.read_from_str(ctx, pattern)
    reorder_proj = reorder_proj.set_tuple_name(isl.dim_type.in_, space).set_tuple_name(
        isl.dim_type.out, space
    )

    return reorder_proj



import copy
import random
from typing import Callable, Generator

import joblib

from accelforge import Spec, util
from accelforge._accelerated_imports import pd
from accelforge.frontend.mapper.metrics import Metrics
from accelforge.frontend.renames import EinsumName, Rank
from accelforge.mapper.FFM import PmappingGroup
from accelforge.mapper.FFM._join_pmappings.compatibility import Compatibility
from accelforge.mapper.FFM._join_pmappings.compress_pmappings import (
    compress_einsum2pmappings,
)
from accelforge.mapper.FFM._join_pmappings.pmapping_dataframe import PmappingDataframe
from accelforge.mapper.FFM._pareto_df.df_convention import col2reservation
from accelforge.mapper.FFM.pmappings import MultiEinsumPmappings
from accelforge.mapper._simanneal2.tracking import EvaluationsScoreTracker
from accelforge.util._frozenset import oset
from accelforge.util.parallel import get_n_parallel_jobs


class FailedMutation(Exception):
    pass


MAX_COMBOS = 100000


def clear_compatibility(
    c: Compatibility | PmappingGroup, shared_tensors: set[str]
) -> Compatibility:
    if isinstance(c, PmappingGroup):
        c = c.compatibility
    return c.clear_dead_tensors(
        shared_tensors
    ).clear_tile_patterns_and_reservation_indices()


class MapspaceGlobals:
    def __init__(
        self,
        einsum2sims: dict[EinsumName, list[PmappingGroup]],
        aliased_tensors: dict[str, set[str]],
        objective_function: Callable[[pd.Series], float],
        tracker: EvaluationsScoreTracker,
        mixable_ranks: dict[Rank, set[Rank]] | None = None,
    ) -> None:
        self.einsum2sims: dict[EinsumName, list[PmappingGroup]] = einsum2sims
        self.aliased_tensors: dict[str, set[str]] = aliased_tensors
        self.objective_function: Callable[[pd.Series], float] = objective_function
        self.tracker: EvaluationsScoreTracker = tracker
        self.pick_lowest_usage_first: bool = False
        self.mixable_ranks: dict[Rank, set[Rank]] | None = mixable_ranks
        self.valid_combos: list[tuple[PmappingGroup, ...]] = []

    def _enumerate_valid_combos(self) -> list[list[tuple[PmappingGroup, ...]]]:
        """Enumerate all valid combinations of compatibility groups across
        einsums. Returns a list of "blocks" — each block covers a contiguous
        range of einsums. When combos exceed MAX_COMBOS, the current block is
        finalized and a new block starts from the next einsum."""
        einsum2sims = self.einsum2sims
        einsums = list(einsum2sims.keys())
        if not einsums:
            return []

        from tqdm import tqdm

        all_blocks = []
        block_start = 0
        combos = [
            (g, g.compatibility)
            for g in einsum2sims[einsums[0]]
            if self.satisfies_mixable_ranks(g.compatibility)
        ]
        print(f"  Einsum {einsums[0]}: {len(combos)} valid groups")

        for i in range(1, len(einsums)):
            e = einsums[i]
            shared_tensors = self._einsum2tensors(
                range(block_start, i)
            ) & self._einsum2tensors(i)
            future_live = (
                self._einsum2tensors(range(i + 1, len(einsums)))
                if i + 1 < len(einsums)
                else oset()
            )
            next_groups = [
                g
                for g in einsum2sims[e]
                if self.satisfies_mixable_ranks(g.compatibility)
            ]

            freed_to_groups = {}
            for g in next_groups:
                freed_to_groups.setdefault(
                    clear_compatibility(g, shared_tensors), []
                ).append(g)

            new_combos = []
            for *groups, joined_compat in tqdm(
                combos, desc=f"  Einsum {e}", leave=False
            ):
                joined_cleared = joined_compat.clear_dead_tensors(
                    future_live | shared_tensors
                )
                shared = joined_cleared.tensor_names & shared_tensors
                if not shared:
                    for g in next_groups:
                        try:
                            new_joined = joined_cleared.merge_next(
                                g.compatibility, future_live
                            )
                        except ValueError:
                            continue
                        new_combos.append((*groups, g, new_joined))
                    continue

                joined_freed = clear_compatibility(joined_cleared, shared)
                for g in freed_to_groups.get(joined_freed, []):
                    try:
                        new_joined = joined_cleared.merge_next(
                            g.compatibility, future_live
                        )
                    except ValueError:
                        continue
                    new_combos.append((*groups, g, new_joined))

            if not new_combos:
                for g in combos:
                    print(f"    LEFT: No match for {g[-1]}")
                for g in next_groups:
                    print(f"    RIGHT: No match for {g.compatibility}")
            combos = new_combos
            print(f"  Einsum {e}: {len(combos)} valid combos")
            assert combos

            # If too many combos, finalize this block and start fresh
            if len(combos) > MAX_COMBOS:
                stripped = [tuple(c[:-1]) for c in combos]
                all_blocks.append((block_start, stripped))
                print(
                    f"  Block {len(all_blocks)}: einsums {block_start}-{i} -> {len(stripped)} combos (capped)"
                )
                # Start new block from this einsum
                block_start = i
                combos = [
                    (g, g.compatibility)
                    for g in einsum2sims[einsums[i]]
                    if self.satisfies_mixable_ranks(g.compatibility)
                ]

        # Finalize last block
        stripped = [tuple(c[:-1]) for c in combos]
        all_blocks.append((block_start, stripped))
        print(
            f"  Block {len(all_blocks)}: einsums {block_start}-{len(einsums)-1} -> {len(stripped)} combos"
        )

        total = sum(len(b[1]) for b in all_blocks)
        print(f"  Total: {len(all_blocks)} blocks, {total} combos")
        return all_blocks

    def _einsum2tensors(
        self, e: EinsumName | int | Generator[EinsumName | int, None, None]
    ) -> set[str]:
        if isinstance(e, Generator) or isinstance(e, range):
            return oset.union(oset(), *(self._einsum2tensors(i) for i in e))
        if isinstance(e, int):
            e = list(self.einsum2sims.keys())[e]
        return self.einsum2sims[e][0].compatibility.tensor_names

    def _filter_einsum2sims(
        self,
        einsum2sims: dict[EinsumName, list[PmappingGroup]],
    ) -> dict[EinsumName, list[PmappingGroup]]:
        """Remove PmappingGroups that have no compatible partner in any
        neighbor einsum. Iterates until no more groups are removed."""
        einsum2sims = {e: list(sims) for e, sims in einsum2sims.items()}

        prev_len = 0
        while sum(len(sims) for sims in einsum2sims.values()) != prev_len:
            prev_len = sum(len(sims) for sims in einsum2sims.values())
            print(f"Filtering einsum2sims. Current length: {prev_len}")

            einsums = list(einsum2sims.keys())
            for i, e in enumerate(einsums):
                for e2 in einsums[i + 1 :]:
                    shared_tensors = self._einsum2tensors(e) & self._einsum2tensors(e2)
                    if not shared_tensors:
                        continue

                    freed_e = {
                        clear_compatibility(s, shared_tensors) for s in einsum2sims[e]
                    }
                    freed_e2 = {
                        clear_compatibility(s, shared_tensors) for s in einsum2sims[e2]
                    }
                    einsum2sims[e] = [
                        s
                        for s in einsum2sims[e]
                        if clear_compatibility(s, shared_tensors) in freed_e2
                    ]
                    einsum2sims[e2] = [
                        s
                        for s in einsum2sims[e2]
                        if clear_compatibility(s, shared_tensors) in freed_e
                    ]

        print(f"Filtered einsum2sims. Final length: {prev_len}")
        return einsum2sims

    def satisfies_mixable_ranks(
        self,
        compatibility: Compatibility,
    ) -> bool:
        if self.mixable_ranks is None:
            return True
        for i in range(len(compatibility.loops)):
            ranks = oset()
            for t in compatibility.tensors:
                if i < len(t.loops):
                    ranks.update(t.loops[i].rank_name)
            for r in ranks:
                if r not in self.mixable_ranks:
                    continue
                for r2 in ranks:
                    if r2 not in self.mixable_ranks[r]:
                        return False
        return True

    def joined_satisfies_mixable_ranks(
        self,
        left: PmappingGroup,
        right: PmappingGroup,
    ) -> bool:
        shared_tensors = (
            left.compatibility.tensor_names & right.compatibility.tensor_names
        )
        joined_compatibility = left.compatibility.merge_next(
            right.compatibility,
            shared_tensors,
        )
        return self.satisfies_mixable_ranks(joined_compatibility)


class SimAnnealMapping:
    def __init__(self, mapspace_globals: MapspaceGlobals) -> None:
        # self.einsum2sim: dict[EinsumName, PmappingGroup] = {
        #     e: random.choice(s) for e, s in mapspace_globals.einsum2sims.items()
        # }
        self.mapspace_globals: MapspaceGlobals = mapspace_globals
        self.einsum2sim: dict[EinsumName, PmappingGroup] = {
            e: random.choice(s) for e, s in mapspace_globals.einsum2sims.items()
        }
        self.einsum2index: dict[EinsumName, int] = {e: 0 for e in self.einsum2sim}
        self.init_choices()
        self._prev_score = None

    def mutate(self) -> None:
        # Pick a random einsum
        e = random.choice(list(self.einsum2sim.keys()))

        random.choice(
            [
                self._randomize_index,
                self._randomize_sim,
            ]
        )(e)
        try:
            self.ensure_match(e)
        except FailedMutation:
            raise
        finally:
            self.add_evaluations()

    def add_evaluations(self) -> None:
        s = sum(sim.mappings.n_total_pmappings for sim in self.einsum2sim.values())
        s *= len(self.mapspace_globals.einsum2sims) / len(self.einsum2sim)
        self.mapspace_globals.tracker.add_evaluation(s, float("inf"))

    def _randomize_index(self, e: EinsumName) -> None:
        self._prev_score = None
        self.einsum2index[e] = random.randint(0, 10000000000000)

    def _randomize_sim(self, e: EinsumName) -> None:
        self.einsum2sim[e] = random.choice(self.mapspace_globals.einsum2sims[e])
        self._randomize_index(e)

    def _einsum_position_in_list(self, e: EinsumName) -> int:
        return list(self.einsum2sim.keys()).index(e)

    def _randomize_lowest_reservation(
        self, e: EinsumName, valid_indices: list[int] | None = None
    ):
        """Pick the row index with the lowest reservations"""
        group = self.einsum2sim[e]
        res_cols = [c for c in group.mappings.data.columns if "reservation" in c]
        if not res_cols:
            self._randomize_index(e)
            return

        data = (
            group.mappings.data
            if valid_indices is None
            else group.mappings.data.iloc[valid_indices]
        )
        self.einsum2index[e] = data[res_cols].max(axis=1).idxmin()

    def _get_valid_right_compatibilities(
        self,
        left: PmappingGroup,
        possible_rights: list[PmappingGroup],
        live_tensors: set[str],
        future_einsums: list[EinsumName],
    ) -> list[PmappingGroup]:
        """Filter possible right PmappingGroups to those whose joined
        compatibility with the left has at least one valid match in every
        future einsum (lookahead)."""
        if not future_einsums:
            return possible_rights

        # Precompute: for each future einsum that shares tensors with left,
        # build the set of freed compatibilities once (not per candidate).
        left_tensors = left.compatibility.tensor_names
        future_checks = []  # list of (shared_tensors, freed_compat_set)
        for future_name in future_einsums:
            future_tensors = self._einsum2tensors(future_name)
            shared = left_tensors & future_tensors
            if not shared:
                continue
            freed = frozenset(
                clear_compatibility(g, shared)
                for g in self.mapspace_globals.einsum2sims[future_name]
            )
            future_checks.append((shared, freed))

        if not future_checks:
            return possible_rights

        result = []
        for right in possible_rights:
            try:
                joined_compat = left.compatibility.merge_next(
                    right.compatibility, live_tensors
                )
            except ValueError:
                continue

            ok = True
            for shared, future_freed in future_checks:
                if clear_compatibility(joined_compat, shared) not in future_freed:
                    ok = False
                    break
            if ok:
                result.append(right)
        return result

    def _get_valid_right_indices(
        self,
        left: PmappingGroup,
        right: PmappingGroup,
        live_tensors: set[str],
        right_tensors: set[str],
        left_row_idx: int | None = None,
    ) -> list[int]:
        """Find which right-side row indices survive the inner join with the
        left side. Tags right with _INDEX, calls PmappingDataframe.merge_next
        directly with ALL resources ignored, then reads surviving _INDEX.

        If left_row_idx is None, looks up the index from self.einsum2index."""
        import copy as _copy

        # Reduce left to just the chosen row
        if left_row_idx is None:
            left_row_idx = self.einsum2index.get(
                next((e for e, v in self.einsum2sim.items() if v is left), None), 0
            )
        left_data = left.mappings.data
        left_row_idx = left_row_idx % len(left_data)
        left_df = PmappingDataframe(
            left_data.iloc[left_row_idx : left_row_idx + 1].copy(),
            n_total_pmappings=1,
            n_valid_pmappings=1,
            ignored_resources=left.mappings.ignored_resources,
            drop_valid_reservations=left.mappings.drop_valid_reservations,
        )

        # Tag right rows with _INDEX
        right_mappings = _copy.copy(right.mappings)
        right_mappings._data = right.mappings.data.assign(
            _INDEX=range(len(right.mappings.data))
        )

        # Collect all resource names so we can ignore them all
        all_resources = oset(
            col.split("<SEP>")[1]
            for df in [left_data, right.mappings.data]
            for col in df.columns
            if col.startswith("reservation<SEP>")
        )

        try:
            joined_compat = left.compatibility.merge_next(
                right.compatibility, live_tensors
            )
            merged = left_df.merge_next(
                right_mappings,
                duplicated_aliased_tensors=oset(),
                compatibility_left=left.compatibility,
                compatibility_right=right.compatibility,
                compatibility_joined=joined_compat,
                ignored_resources=all_resources,
            )
            if "_INDEX" in merged.data.columns:
                return list(oset(merged.data["_INDEX"]))
            return list(range(len(right.mappings.data)))
        except (ValueError, FailedMutation):
            return []

    def init_choices(self) -> None:
        if not self.mapspace_globals.valid_combos:
            raise FailedMutation("No valid compatibility combinations exist")

        all_einsums = list(self.mapspace_globals.einsum2sims.keys())
        self.einsum2sim = {}

        # Pick one combo from each block
        for block_start, block_combos in self.mapspace_globals.valid_combos:
            combo = random.choice(block_combos)
            block_einsums = all_einsums[block_start : block_start + len(combo)]
            for einsum_name, group in zip(block_einsums, combo):
                self.einsum2sim[einsum_name] = group

        # Pick row indices — use _get_valid_right_indices to pre-filter
        for idx, einsum_name in enumerate(all_einsums):
            if self.mapspace_globals.pick_lowest_usage_first:
                self._randomize_lowest_reservation(einsum_name)
            else:
                self._randomize_index(einsum_name)

            if idx > 0:
                prev_name = all_einsums[idx - 1]
                right_tensors = self._einsum2tensors(prev_name)
                live_tensors = (
                    self._einsum2tensors(range(idx, len(all_einsums)))
                    if idx < len(all_einsums)
                    else oset()
                )
                valid = self._get_valid_right_indices(
                    self.einsum2sim[prev_name],
                    self.einsum2sim[einsum_name],
                    live_tensors,
                    right_tensors,
                )
                if valid:
                    if self.mapspace_globals.pick_lowest_usage_first:
                        self._randomize_lowest_reservation(einsum_name, valid)
                    else:
                        self.einsum2index[einsum_name] = random.choice(valid)

        self.add_evaluations()

    def ensure_match(
        self,
        lock_choice_for_einsum: EinsumName,
    ) -> None:

        new_einsum2sim: dict[EinsumName, PmappingGroup] = {}

        # Grab all the compatibilities that match
        for i, (e, s) in enumerate[tuple[EinsumName, PmappingGroup]](
            list(self.einsum2sim.items())
        ):
            if e == lock_choice_for_einsum:
                new_einsum2sim[e] = s
                continue

            following_tensors = self._einsum2tensors(range(i + 1, len(self.einsum2sim)))

            to_check = [(s2, s) for s2 in new_einsum2sim.values()]

            if i < self._einsum_position_in_list(lock_choice_for_einsum):
                to_check.append((s, self.einsum2sim[lock_choice_for_einsum]))
            else:
                to_check.append((self.einsum2sim[lock_choice_for_einsum], s))

            for left, right in to_check:
                shared = (
                    left.compatibility.tensor_names & right.compatibility.tensor_names
                )
                if clear_compatibility(left, shared) != clear_compatibility(
                    right, shared
                ):
                    break
                cl = clear_compatibility(left, following_tensors)
                cr = clear_compatibility(right, following_tensors)
                if cl.n_loops > cr.n_loops:
                    break

            else:
                new_einsum2sim[e] = s

        # Grab compatibilities that don't match
        def _matches(s: PmappingGroup, c: Compatibility) -> bool:
            shared = s.compatibility.tensor_names & c.tensor_names
            return clear_compatibility(s, shared) == clear_compatibility(c, shared)

        for e, pmapping_groups in self.mapspace_globals.einsum2sims.items():
            if e in new_einsum2sim:
                continue

            for s in new_einsum2sim.values():
                pmapping_groups = [
                    s2 for s2 in pmapping_groups if _matches(s2, s.compatibility)
                ]

            if not pmapping_groups:
                # print(f"No compatible PmappingGroups found for {e}")
                raise FailedMutation(f"No compatible PmappingGroups found for {e}")

            new_einsum2sim[e] = random.choice(pmapping_groups)
            self._randomize_index(e)

            # pmapping_groups = self.mapspace_globals.einsum2sims[e]
            # [s.compatibility for s in self.einsum2sim.values()]
            # [s.compatibility for s in new_einsum2sim.values()]
            # {e: s.compatibility for e, s in new_einsum2sim.items()}

        assert len(new_einsum2sim) == len(self.einsum2sim)
        assert oset(new_einsum2sim.keys()) == oset(self.einsum2sim.keys())
        self.einsum2sim = {k: new_einsum2sim[k] for k in self.einsum2sim.keys()}

    def _einsum2tensors(
        self, e: EinsumName | int | Generator[EinsumName | int, None, None]
    ) -> set[str]:
        return self.mapspace_globals._einsum2tensors(e)

    def _access_index(self, e: EinsumName, index_override: int | None = None):
        s = self.einsum2sim[e]
        data = s.mappings.data
        i = self.einsum2index[e] if index_override is None else index_override
        i %= len(data)
        return PmappingGroup(
            compatibility=s.compatibility,
            mappings=PmappingDataframe(
                data.iloc[i : i + 1].copy(),
                n_total_pmappings=s.mappings.n_total_pmappings,
                n_valid_pmappings=s.mappings.n_valid_pmappings,
                ignored_resources=s.mappings.ignored_resources,
                drop_valid_reservations=s.mappings.drop_valid_reservations,
            ),
        )

    def get_score(self) -> float:
        if self._prev_score is not None:
            return self._prev_score

        all_einsum_names = list(self.einsum2sim.keys())
        items: list[tuple[EinsumName, PmappingGroup]] = list(self.einsum2sim.items())
        joined: PmappingGroup = items.pop(0)[1]
        joined = copy.deepcopy(joined)
        for i, (e, s) in enumerate(items):
            # We're starting from 1, so i in items means einsum index i+1 in the full
            # list
            full_idx = i + 1
            right_tensors = self._einsum2tensors(full_idx)
            live_tensors = (
                self._einsum2tensors(range(full_idx + 1, len(all_einsum_names)))
                if full_idx + 1 < len(all_einsum_names)
                else oset()
            )

            joined.compatibility = joined.compatibility.clear_dead_tensors(
                live_tensors | right_tensors
            )

            def _merge_next(
                left: PmappingGroup,
                right: PmappingGroup,
            ) -> PmappingGroup:
                try:
                    return left.merge_next(
                        right,
                        live_tensors_post_join=live_tensors,
                        live_tensors_with_right=live_tensors | right_tensors,
                        aliased_tensors=self.mapspace_globals.aliased_tensors,
                        compatibility_joined=joined.compatibility.merge_next(
                            s.compatibility,
                            live_tensors,
                        ),
                        permuted_compatibility_left=left.compatibility,
                        permuted_compatibility_right=right.compatibility,
                        delay=False,
                        ignored_resources=oset(),
                    )
                except (ValueError, AssertionError) as err:
                    raise FailedMutation(f"No valid pmappings: {err}")

            # Try to merge using the index we already have set
            joined_new = _merge_next(joined, self._access_index(e))
            if len(joined_new.mappings.data) >= 1:
                # If multiple valid joined mappings, keep only the first one
                if len(joined_new.mappings.data) > 1:
                    joined_new.mappings._data = joined_new.mappings.data.iloc[:1].copy()
                joined = joined_new
                continue

            # No valid pmappings with current index. Use a cheap inner join
            # to find which right-side indices are valid, then pick one.
            self.add_evaluations()

            valid_indices = self._get_valid_right_indices(
                joined,
                self.einsum2sim[e],
                live_tensors,
                right_tensors,
                left_row_idx=0,
            )
            if valid_indices:
                i = random.choice(valid_indices)
                joined_new = _merge_next(joined, self._access_index(e, i))
                if len(joined_new.mappings.data) >= 1:
                    self.einsum2index[e] = i
                    if len(joined_new.mappings.data) > 1:
                        joined_new.mappings._data = joined_new.mappings.data.iloc[
                            :1
                        ].copy()
                    joined = joined_new
                    continue

            # Full fallback: merge all rows and pick a valid one
            s = self.einsum2sim[e]
            s.mappings.data["_INDEX"] = list(range(len(s.mappings.data)))
            joined_new = _merge_next(
                joined,
                s,
            )
            s.mappings._data = s.mappings.data.drop(columns=["_INDEX"])
            try:
                i = random.choice(list(oset(joined_new.mappings.data["_INDEX"])))
            except IndexError:
                raise FailedMutation(f"No valid pmappings for {e}")

            # Now that we've picked, merge with the index we just set
            joined_new = _merge_next(joined, self._access_index(e, i))

            if len(joined_new.mappings.data) >= 1:
                # If multiple valid joined mappings, keep only the first one
                if len(joined_new.mappings.data) > 1:
                    joined_new.mappings._data = joined_new.mappings.data.iloc[:1].copy()
                # If it worked, set the index
                self.einsum2index[e] = i
                joined = joined_new
                continue

            raise FailedMutation(
                f"Got {len(joined_new.mappings.data)} pmappings for {e}"
            )

        assert len(joined.mappings.data) == 1
        score = self.mapspace_globals.objective_function(joined.mappings.data.iloc[0])
        self.mapspace_globals.tracker.add_evaluation(0, score)
        self._prev_score = score
        return score

    def crossover(self, other: "SimAnnealMapping") -> "SimAnnealMapping":
        child = other.copy()
        e = random.choice(list(self.einsum2sim.keys()))
        child.einsum2sim[e] = self.einsum2sim[e]
        child.einsum2index[e] = self.einsum2index[e]

        try:
            child.ensure_match(e)
        except FailedMutation:
            child = other.copy()

        child.add_evaluations()
        return child

    def copy(self) -> "SimAnnealMapping":
        s = SimAnnealMapping.__new__(SimAnnealMapping)
        s.mapspace_globals = self.mapspace_globals
        s.einsum2sim = self.einsum2sim.copy()
        s.einsum2index = self.einsum2index.copy()
        s._prev_score = self._prev_score
        return s


def get_random_mapping(mapspace_globals: MapspaceGlobals) -> SimAnnealMapping:
    while True:
        try:
            s = SimAnnealMapping(mapspace_globals)
            s.get_score()
            return s
        except FailedMutation:
            if mapspace_globals.tracker.finished():
                return None
            continue


def _make_mapspace_globals(
    pmapping_groups: dict[EinsumName, list[PmappingGroup]],
    spec: Spec,
    tracker: EvaluationsScoreTracker,
    pick_lowest_usage_first: bool = False,
    valid_combos: list[tuple[PmappingGroup, ...]] | None = None,
) -> MapspaceGlobals:
    objective = spec.mapper.metrics
    if objective == Metrics.ENERGY:
        objective_function = lambda x: x["Total<SEP>energy"]
    elif objective == Metrics.LATENCY:
        objective_function = lambda x: x["Total<SEP>latency"]
    elif objective == (Metrics.ENERGY | Metrics.LATENCY):
        objective_function = lambda x: x["Total<SEP>energy"] * x["Total<SEP>latency"]
    else:
        raise ValueError(f"Unknown objective {objective}")
    mg = MapspaceGlobals(
        einsum2sims=pmapping_groups,
        aliased_tensors=spec.workload.get_tensor_copies(),
        objective_function=objective_function,
        tracker=tracker,
    )
    mg.pick_lowest_usage_first = pick_lowest_usage_first
    if valid_combos is not None:
        mg.valid_combos = valid_combos
    return mg


def _make_initial_population(
    mapspace_globals: MapspaceGlobals,
    pop_size: int,
) -> list[SimAnnealMapping]:
    tracker = mapspace_globals.tracker
    mappings = []
    while len(mappings) < pop_size:
        m = get_random_mapping(mapspace_globals)
        if m is None:
            return mappings
        mappings.append(m)
        if tracker.finished():
            print(f"Timed out after {len(mappings)}/{pop_size} mappings.")
            return mappings
    print(
        f"Completed making initial population of {len(mappings)} mappings after {tracker.evaluations}  evaluations."
    )
    return mappings


def _join_pmappings(
    pmapping_groups: dict[EinsumName, list[PmappingGroup]],
    spec: Spec,
    tracker: EvaluationsScoreTracker,
    pop_size_per_thread: int,
    algorithm: str = "simanneal",
    pick_lowest_usage_first: bool = False,
    valid_combos: list[tuple[PmappingGroup, ...]] | None = None,
) -> PmappingGroup:
    mapspace_globals = _make_mapspace_globals(
        pmapping_groups, spec, tracker, pick_lowest_usage_first, valid_combos
    )
    mappings = _make_initial_population(mapspace_globals, pop_size_per_thread)
    if not mappings or tracker.finished():
        return

    if algorithm == "simanneal":
        _run_simanneal(mappings, tracker)
    elif algorithm == "genetic":
        _run_genetic(mappings, tracker, pop_size_per_thread)
    else:
        raise ValueError(f"Unknown algorithm: {algorithm}")


def _run_simanneal(
    mappings: list[SimAnnealMapping],
    tracker: EvaluationsScoreTracker,
) -> None:
    while not tracker.finished():
        for i, m in enumerate(list(mappings)):
            try:
                new = m.copy()
                new.mutate()
                if new.get_score() < m.get_score():
                    mappings[i] = new
                if tracker.finished():
                    break
            except FailedMutation:
                continue


def _run_genetic(
    mappings: list[SimAnnealMapping],
    tracker: EvaluationsScoreTracker,
    pop_size: int,
    crossover_rate: float = 0.7,
    mutation_rate: float = 0.3,
) -> None:
    while not tracker.finished():
        fitness = []
        for m in mappings:
            try:
                fitness.append(m.get_score())
            except FailedMutation:
                fitness.append(float("inf"))

        best_idx = min(range(len(fitness)), key=lambda i: fitness[i])
        best_mapping = mappings[best_idx]

        inv_fitness = [1.0 / (f + 1e-9) for f in fitness]
        total = sum(inv_fitness)
        n = len(inv_fitness)
        probs = [f / total for f in inv_fitness] if total > 0 else [1.0 / n] * n
        selected = random.choices(range(len(mappings)), probs, k=pop_size)

        new_pop: list[SimAnnealMapping] = []
        for i in range(0, len(selected), 2):
            p1 = mappings[selected[i]]
            p2 = mappings[selected[(i + 1) % len(selected)]]
            if random.random() < crossover_rate:
                new_pop.append(p1.crossover(p2))
                new_pop.append(p2.crossover(p1))
            else:
                new_pop.append(p1.copy())
                new_pop.append(p2.copy())

        for m in new_pop:
            if random.random() < mutation_rate:
                try:
                    m.mutate()
                except FailedMutation:
                    pass

        new_pop.append(best_mapping)
        mappings = new_pop[:pop_size]

        if tracker.finished():
            break


def get_n_tile_shapes(sim: PmappingGroup) -> int:
    df = sim.mappings.data
    symbols = sim.compatibility.symbols()
    cols = [c for c in df.columns if c in symbols]
    if not cols:
        return 1
    return len(df.groupby(cols).size())


def join_pmappings(
    pmappings: MultiEinsumPmappings,
    max_evaluations: int = 1,
    population_size=1000,
    score_target: float | None = None,
    algorithm: str = "simanneal",
    optimal: float | None = None,
    pick_lowest_usage_first: bool = False,
) -> EvaluationsScoreTracker:

    spec = pmappings.spec
    tracker = EvaluationsScoreTracker(
        max_evaluations=max_evaluations / get_n_parallel_jobs(),
        stop_at_score=None,
        print_period=1,
        optimal=optimal,
    )

    # Disable validation in the compatibility class to avoid errors when joining
    # pmappings. We use a simplified version of the joining process in this code that
    # doesn't do all the reservation cleaning that we do in FFM.
    Compatibility.__post_init__ = lambda *args, **kwargs: None

    compressed, decompress_data = compress_einsum2pmappings(pmappings.einsum2pmappings)

    if population_size != float("inf"):
        pop_size_per_thread = max(1, population_size // get_n_parallel_jobs())
    else:
        pop_size_per_thread = float("inf")

    # Multiply by the number of einsums
    # print(
    #     f"Multiplying scale by {len(pmappings.einsum2pmappings)} for number of einsums"
    # )
    # tracker._scale_by *= len(pmappings.einsum2pmappings)

    # We allow FFM to query n_pareto_optimal_pmappings mappings from the mapspace, so we
    # scale by 1 / n_pareto_optimal_pmappings to get simanneal 1 evaluation = same
    # number of mappings as FFM
    # print(
    #     f"Multiplying scale by {1 / pmappings.n_pareto_optimal_pmappings()} for Pareto-optimal mapspace size"
    # )
    # tracker._scale_by *= 1 / pmappings.n_pareto_optimal_pmappings()

    for einsum_name, einsum_pmappings in pmappings.einsum2pmappings.items():
        total = sum(len(p.mappings.data) for p in einsum_pmappings)
        n_compatibilities = len(einsum_pmappings)
        print(
            f"Einsum {einsum_name} has {total} pmappings with {n_compatibilities} compatibilities"
        )
        if total == 0:
            raise ValueError(f"Einsum {einsum_name} has no pmappings")

    permuted = {}
    n_evaluated = 1
    for einsum_name, einsum_sims in compressed.items():
        for s in einsum_sims:
            n_pmappings = s.mappings.n_total_pmappings
            n_evaluated += n_pmappings
            # Count all permutations as separate choices
            for c_perm, _ in s.compatibility.make_equivalent_permutations():
                permuted.setdefault(einsum_name, []).append(
                    PmappingGroup(
                        compatibility=c_perm,
                        mappings=s.mappings,
                    )
                )

    print(f"Multiplying scale by {1 / n_evaluated} for number of evaluated pmappings")
    tracker._scale_by *= 1 / n_evaluated

    mg = MapspaceGlobals(
        permuted,
        {},
        lambda x: 0,
        tracker,
        mixable_ranks=spec.workload._get_ranks_that_share_indexing_rank_variables(),
    )
    permuted = mg._filter_einsum2sims(permuted)
    mg.einsum2sims = permuted
    mg.einsum2sims = permuted
    valid_combos = mg._enumerate_valid_combos()

    def parallel_join(
        permuted: dict[EinsumName, list[PmappingGroup]],
        spec: Spec,
        tracker: EvaluationsScoreTracker,
        pop_size_per_thread: int,
        algorithm: str,
        pick_lowest_usage_first: bool,
        valid_combos: list[tuple[PmappingGroup, ...]],
    ) -> EvaluationsScoreTracker:
        _join_pmappings(
            permuted,
            spec,
            tracker,
            pop_size_per_thread,
            algorithm,
            pick_lowest_usage_first,
            valid_combos,
        )
        return tracker

    trackers = util.parallel(
        joblib.delayed(parallel_join)(
            permuted,
            spec,
            tracker,
            pop_size_per_thread,
            algorithm,
            pick_lowest_usage_first,
            valid_combos,
        )
        for _ in range(get_n_parallel_jobs())
    )

    t0 = trackers[0]
    for t in trackers[1:]:
        t0.merge_with(t)

    # for einsum_name in pmappings.einsum2pmappings:
    #     col = f"{einsum_name}<SEP>{MAPPING_COLUMN}"
    #     joined.data[col] = joined.data[col].apply(
    #         lambda x: pmappings.pmapping_objects[einsum_name][x]
    #     )

    # rank_variable_bounds = get_rank_variable_bounds_for_all_einsums(spec)
    # joined.data[f"Total<SEP>{MAPPING_COLUMN}"] = joined.data.apply(
    #     lambda row: MappingFromRow(row, spec, rank_variable_bounds), axis=1
    # )
    # # Fill nans with 0. We might get missing columns for some mapping entries if there
    # # are energy entries for some pmappings but not others (e.g., one pmapping accesses
    # # DRAM while another doesn't.)
    # joined._data = joined.data._fillna_and__numeric_cast(0)
    return t0  # Mappings(spec, list(pmappings.einsum2pmappings.keys()), joined.data)



from collections.abc import Iterable

import matplotlib.axes as axes
import matplotlib.pyplot as plt

from accelforge.frontend.spec import Spec


def plot_area(
    specs: Iterable[Spec], labels: Iterable[str] = None, ax: axes.Axes = None
):
    """
    Plot area of one or more specs.

    Parameters
    ----------
    specs:
        An iterable of specifications.
    labels:
        An iterable of the same length as `specs` to use as labels in the plot.
    ax:
        An matplotlib Axes to use. A new one is created by default.
    """
    if ax is None:
        fig, ax = plt.subplots()
    ax.set_ylabel("Area (m^2)")

    if labels is None:
        labels = [f"spec-{i}" for i in range(len(specs))]
    assert len(labels) == len(specs)

    component2color = {}
    for i, (label, spec) in enumerate(zip(labels, specs)):
        heights = []
        colors = []
        names = []
        height = 0
        for component, area in spec.arch.per_component_total_area.items():
            height += area
            if component not in component2color:
                color = plt.cm.tab10(len(component2color))
                component2color[component] = color
            else:
                color = component2color[component]
            heights.append(height)
            colors.append(color)
            names.append(component)

        heights = reversed(heights)
        colors = reversed(colors)
        names = reversed(names)
        for height, color, name in zip(heights, colors, names):
            ax.bar(i, height=height, label=name, color=color)

    ax.set_xticks(range(len(specs)), labels)

    ax.legend()



import pydot


def _pydot_graph() -> pydot.Dot:
    graph = pydot.Dot(graph_type="graph", rankdir="TD", ranksep=0.1)
    graph.set_node_defaults(shape="box", fontname="Arial", fontsize="12")
    graph.set_edge_defaults(fontname="Arial", fontsize="10")
    return graph


# =============================================================================
# Color Map for Visualization
# =============================================================================


class ColorMap:

    def __init__(self, keys: list[str]):
        self.keys = keys
        self.color_list = self._make_color_map(len(keys))
        self.color_map = {key: self.color_list[i] for i, key in enumerate(keys)}

    def format_list(self, items: list[str]) -> str:
        result = ['<<TABLE BORDER="0" CELLBORDER="0" CELLSPACING="0"><TR>']
        for i, item in enumerate(items):
            start = '<TD ALIGN="CENTER">'  # if i < len(items) - 1 else f'</TR><TR><TD ALIGN="CENTER" COLSPAN="100">'
            if item in self.color_map:
                start = f'<TD ALIGN="CENTER" BORDER="5" COLOR="{self.color_map[item]}">'
            end = "</TD>"
            result.append(f"{start}{item}{end}")
        result.append("</TR></TABLE>>")
        return "".join(result)

        # This makes a colored bar under the text
        # result = ['<<TABLE BORDER="0" CELLBORDER="0" CELLSPACING="0" CELLPADDING="0">']
        # # First row: text
        # result.append('<TR>')
        # for item in items:
        #     result.append(f'<TD ALIGN="CENTER" STYLE="margin:0;padding:0;">{item}</TD>')
        # result.append('</TR>')
        # # Second row: color bar (height 20, width 40, minimal spacing)
        # result.append('<TR>')
        # for item in items:
        #     if item in self.color_map:
        #         result.append(f'<TD BGCOLOR="{self.color_map[item]}" HEIGHT="10" WIDTH="15" FIXEDSIZE="TRUE" STYLE="margin:0;padding:0;"></TD>')
        #     else:
        #         result.append('<TD HEIGHT="20" WIDTH="40" FIXEDSIZE="TRUE" STYLE="margin:0;padding:0;"></TD>')
        # result.append('</TR>')
        # result.append('</TABLE>>')
        # return ''.join(result)

    def _make_color_map(self, n_colors: int) -> list[str]:
        if n_colors <= 0:
            return []

        # High contrast, distinguishable colors for borders
        base_colors = [
            "#FF0000",  # Red
            "#00FF00",  # Green
            "#0000FF",  # Blue
            "#FFFF00",  # Yellow
            "#FF00FF",  # Magenta
            "#00FFFF",  # Cyan
            "#FF8000",  # Orange
            "#8000FF",  # Purple
            "#008000",  # Dark Green
            "#800000",  # Dark Red
            "#000080",  # Dark Blue
            "#808000",  # Olive
        ]

        if n_colors <= len(base_colors):
            return base_colors[:n_colors]

        # For more colors, generate additional colors with maximum distinction
        colors = base_colors.copy()

        # Use evenly spaced hues for maximum distinction
        for i in range(len(base_colors), n_colors):
            # Evenly space hues around the color wheel
            hue = i / n_colors

            # Use high saturation and value for maximum contrast
            saturation = 1.0  # Full saturation
            value = 1.0  # Full value

            # Convert HSV to RGB
            h = hue * 6
            c = value * saturation
            x = c * (1 - abs(h % 2 - 1))
            m = value - c

            if h < 1:
                r, g, b = c, x, 0
            elif h < 2:
                r, g, b = x, c, 0
            elif h < 3:
                r, g, b = 0, c, x
            elif h < 4:
                r, g, b = 0, x, c
            elif h < 5:
                r, g, b = x, 0, c
            else:
                r, g, b = c, 0, x

            r = int((r + m) * 255)
            g = int((g + m) * 255)
            b = int((b + m) * 255)

            colors.append(f"#{r:02x}{g:02x}{b:02x}")

        return colors



from typing import Annotated

from pydantic import ConfigDict
from accelforge.util._basetypes import EvalableModel, EvalExtras, EvalsTo


class Variables(EvalExtras):
    """
    Variables that can be used in parsing. All variables defined here can be referenced
    elsewhere in any of the Spec's evaluated expressions.
    """



"""
Regression tests for the FFM mapper.

    python tests/test_regression.py            # (re)generate both reference jsons
    pytest tests/test_regression.py -v         # compare against reference
"""

import json
import math
from numbers import Number
import os
import re
import unittest
from pathlib import Path

import accelforge as af
from accelforge.frontend.spec import Spec
from accelforge.mapper import Metrics

JSON_PATH = Path(__file__).parent / "regression_reference.json"
HWCOMPONENTS_JSON_PATH = Path(__file__).parent / "hwcomponents_expected.json"

REGRESSION_CASES = {
    af.examples.arches.simple: [
        {
            "workload": af.examples.workloads.matmuls,
            "jinja_parse_data": {"N_EINSUMS": 2, "M": 64, "KN": 64},
        },
        {"workload": af.examples.workloads.three_matmuls_annotated},
        {"workload": af.examples.workloads.gpt3_175B},
        {"workload": af.examples.workloads.gpt3_175B_kv_cache},
        # {"workload": af.examples.workloads.gpt3_6_7B},
        # {"workload": af.examples.workloads.gpt3_6_7B_kv_cache},
    ],
    af.examples.arches.eyeriss: [
        {
            "workload": af.examples.workloads.matmuls,
            "jinja_parse_data": {"N_EINSUMS": 2, "M": 64, "KN": 64},
        },
        {"workload": af.examples.workloads.three_matmuls_annotated},
        # {"workload": af.examples.workloads.gpt3_175B},
        # {"workload": af.examples.workloads.gpt3_175B_kv_cache},
        {"workload": af.examples.workloads.gpt3_6_7B},
        {"workload": af.examples.workloads.gpt3_6_7B_kv_cache},
    ],
    af.examples.arches.simba: [
        {
            "workload": af.examples.workloads.matmuls,
            "jinja_parse_data": {"N_EINSUMS": 2, "M": 64, "KN": 64},
        },
        {"workload": af.examples.workloads.three_matmuls_annotated},
        # {"workload": af.examples.workloads.gpt3_175B},
        # {"workload": af.examples.workloads.gpt3_175B_kv_cache},
        {"workload": af.examples.workloads.gpt3_6_7B},
        {"workload": af.examples.workloads.gpt3_6_7B_kv_cache},
    ],
    af.examples.arches.tpu_v4i: [
        {
            "workload": af.examples.workloads.matmuls,
            "jinja_parse_data": {"N_EINSUMS": 2, "M": 64, "KN": 64},
        },
        {"workload": af.examples.workloads.three_matmuls_annotated},
        {"workload": af.examples.workloads.gpt3_175B},
        {"workload": af.examples.workloads.gpt3_175B_kv_cache},
        # {"workload": af.examples.workloads.gpt3_6_7B},
        # {"workload": af.examples.workloads.gpt3_6_7B_kv_cache},
    ],
}


def cast(d):
    if isinstance(d, dict):
        return {str(k): cast(v) for k, v in d.items()}
    if isinstance(d, list):
        return [cast(v) for v in d]
    if isinstance(d, Number):
        return float(d)
    return d


def _key(arch, workload, fused):
    jinja = workload.get("jinja_parse_data", {})
    j = ",".join(f"{k}={v}" for k, v in sorted(jinja.items()))
    return (
        f"{Path(arch).stem}|{Path(workload['workload']).stem}"
        f"|{j}|{'fused' if fused else 'unfused'}"
    )


def _run(arch, workload, fused, print_progress: bool = True):
    spec = Spec.from_yaml(
        arch,
        workload["workload"],
        jinja_parse_data=workload.get("jinja_parse_data"),
    )
    spec.mapper.metrics = Metrics.ENERGY
    spec.mapper.max_fused_loops = 1
    if not fused:
        for node in spec.arch.nodes:
            if isinstance(node, af.arch.Memory):
                node.tensors.keep = "All"
                break
    mappings = spec.map_workload_to_arch(print_progress=print_progress)
    m = mappings[0]
    return cast(
        {
            "energy": float(m.energy()),
            "latency": float(m.latency()),
            "energy_per_component": m.energy(
                per_component=True, per_einsum=True, per_action=True
            ),
            "latency_per_component": m.latency(per_component=True, per_einsum=True),
            "actions": m.actions(per_component=True, per_einsum=True, per_tensor=True),
            "n_mappings": int(len(mappings)),
        }
    )


def _cases(fusion_choices=(False, True)):
    for arch, workloads in REGRESSION_CASES.items():
        for workload in workloads:
            for fused in fusion_choices:
                yield _key(arch, workload, fused), arch, workload, fused


def generate(fusion_choices=(False, True)):
    PARALLEL_GENERATE = False
    from accelforge.util import parallel, delayed, get_n_parallel_jobs

    jobs = {
        key: delayed(_run)(arch, workload, fused, print_progress=not PARALLEL_GENERATE)
        for key, arch, workload, fused in _cases(fusion_choices)
    }
    n_jobs = 1 if not PARALLEL_GENERATE else get_n_parallel_jobs()
    results = parallel(jobs, pbar="Generating regression reference", n_jobs=n_jobs)
    with open(JSON_PATH, "w") as f:
        json.dump(results, f, indent=2)
    print(f"Wrote {len(results)} results to {JSON_PATH}")


class TestFFMRegression(unittest.TestCase):
    _ref = None

    @classmethod
    def setUpClass(cls):
        af.set_n_parallel_jobs(os.cpu_count(), print_message=True)
        assert JSON_PATH.exists(), f"No reference json. Run: python {__file__}"
        with open(JSON_PATH) as f:
            cls._ref = json.load(f)

    def _check(self, key, arch, workload, fused):
        if key not in self._ref:
            self.skipTest(f"not in json: {key}")
        ref = self._ref[key]
        cur = _run(arch, workload, fused)
        self.assertEqual(
            cur["n_mappings"],
            ref["n_mappings"],
            msg=f"n_mappings for {key}: ref={ref['n_mappings']}, cur={cur['n_mappings']}",
        )
        self.assertTrue(
            math.isclose(cur["energy"], ref["energy"], rel_tol=0.01),
            msg=f"energy for {key}: ref={ref['energy']}, cur={cur['energy']}",
        )
        self.assertTrue(
            math.isclose(cur["latency"], ref["latency"], rel_tol=0.01),
            msg=f"latency for {key}: ref={ref['latency']}, cur={cur['latency']}",
        )
        failures = []
        for s in ["energy_per_component", "latency_per_component", "actions"]:
            for c in ref[s]:
                if c not in cur[s]:
                    failures.append(
                        f"{s} {c}: missing in current (keys: {sorted(cur[s].keys())})"
                    )
                    continue
                if not math.isclose(cur[s][c], ref[s][c], rel_tol=0.01):
                    ratio = cur[s][c] / ref[s][c] if ref[s][c] != 0 else float("inf")
                    failures.append(
                        f"{s} {c}: ref={ref[s][c]:.6e} cur={cur[s][c]:.6e} "
                        f"ratio={ratio:.4f}"
                    )
        if failures:
            self.fail(
                f"Regression failures for {key} "
                f"(n_mappings ref={ref['n_mappings']} cur={cur['n_mappings']}, "
                f"energy ref={ref['energy']:.6e} cur={cur['energy']:.6e}, "
                f"latency ref={ref['latency']:.6e} cur={cur['latency']:.6e}):\n"
                + "\n".join(f"  {f}" for f in failures)
            )


for _k, _a, _w, _f in _cases():
    _name = "test_" + re.sub(r"[^a-zA-Z0-9]", "_", _k)

    def _t(_k=_k, _a=_a, _w=_w, _f=_f):
        def t(self):
            self._check(_k, _a, _w, _f)

        return t

    setattr(TestFFMRegression, _name, _t())


class TestHWComponentsConsistency(unittest.TestCase):
    """
    Checks that hwcomponents models produce expected per-component area, energy, leak
    power, and latency.

    If these fail, then the other regression tests will likely fail as well.
    """

    _expected = None
    _results = {}

    @classmethod
    def setUpClass(cls):
        assert (
            HWCOMPONENTS_JSON_PATH.exists()
        ), f"No hwcomponents reference json at {HWCOMPONENTS_JSON_PATH}"
        with open(HWCOMPONENTS_JSON_PATH) as f:
            cls._expected = json.load(f)

        arches = {
            "eyeriss": af.examples.arches.eyeriss,
            "simba": af.examples.arches.simba,
            "simple": af.examples.arches.simple,
            "tpu_v4i": af.examples.arches.tpu_v4i,
        }
        for name, arch_path in arches.items():
            spec = Spec.from_yaml(
                arch_path,
                af.examples.workloads.matmuls,
                jinja_parse_data={"N_EINSUMS": 2, "M": 64, "KN": 64},
            )
            spec = spec.calculate_component_area_energy_latency_leak(
                einsum_name="Matmul0"
            )
            components = {}
            for node in spec.arch.nodes:
                if not isinstance(node, (af.arch.Memory, af.arch.Compute)):
                    continue
                comp = {}
                if node.area is not None:
                    comp["area"] = float(node.area)
                if node.leak_power is not None:
                    comp["leak_power"] = float(node.leak_power)
                actions = {}
                for a in node.actions:
                    act = {}
                    if a.energy is not None:
                        act["energy"] = float(a.energy)
                    if a.latency is not None:
                        act["latency"] = float(a.latency)
                    if act:
                        actions[a.name] = act
                if actions:
                    comp["actions"] = actions
                if comp:
                    components[node.name] = comp
            cls._results[name] = components

    def _check_arch(self, name):
        expected = self._expected[name]
        actual = self._results[name]
        for comp_name, exp_comp in expected.items():
            self.assertIn(comp_name, actual, f"{name}: missing component {comp_name}")
            act_comp = actual[comp_name]
            for field in ["area", "leak_power"]:
                if field in exp_comp:
                    exp_val = exp_comp[field]
                    act_val = act_comp.get(field)
                    self.assertIsNotNone(act_val, f"{name}.{comp_name}.{field} missing")
                    delta = max(abs(exp_val) * 1e-6, 1e-15)
                    self.assertAlmostEqual(
                        act_val,
                        exp_val,
                        delta=delta,
                        msg=f"{name}.{comp_name}.{field}: {act_val} != {exp_val}",
                    )
            for act_name, exp_act in exp_comp.get("actions", {}).items():
                self.assertIn(
                    act_name,
                    act_comp.get("actions", {}),
                    f"{name}.{comp_name}: missing action {act_name}",
                )
                act_act = act_comp["actions"][act_name]
                for field, exp_val in exp_act.items():
                    act_val = act_act.get(field)
                    self.assertIsNotNone(
                        act_val, f"{name}.{comp_name}.{act_name}.{field} missing"
                    )
                    delta = max(abs(exp_val) * 1e-6, 1e-15)
                    self.assertAlmostEqual(
                        act_val,
                        exp_val,
                        delta=delta,
                        msg=f"{name}.{comp_name}.{act_name}.{field}: {act_val} != {exp_val}",
                    )


for _arch_name in ["eyeriss", "simba", "simple", "tpu_v4i"]:
    _test_name = f"test_{_arch_name}"

    def _make_test(_name=_arch_name):
        def t(self):
            self._check_arch(_name)

        return t

    setattr(TestHWComponentsConsistency, _test_name, _make_test())


def generate_hwcomponents():
    arches = {
        "eyeriss": af.examples.arches.eyeriss,
        "simba": af.examples.arches.simba,
        "simple": af.examples.arches.simple,
        "tpu_v4i": af.examples.arches.tpu_v4i,
    }
    results = {}
    for name, arch_path in arches.items():
        spec = Spec.from_yaml(
            arch_path,
            af.examples.workloads.matmuls,
            jinja_parse_data={"N_EINSUMS": 2, "M": 64, "KN": 64},
        )
        spec = spec.calculate_component_area_energy_latency_leak(einsum_name="Matmul0")
        components = {}
        for node in spec.arch.nodes:
            if not isinstance(node, (af.arch.Memory, af.arch.Compute)):
                continue
            comp = {}
            if node.area is not None:
                comp["area"] = float(node.area)
            if node.leak_power is not None:
                comp["leak_power"] = float(node.leak_power)
            actions = {}
            for a in node.actions:
                act = {}
                if a.energy is not None:
                    act["energy"] = float(a.energy)
                if a.latency is not None:
                    act["latency"] = float(a.latency)
                if act:
                    actions[a.name] = act
            if actions:
                comp["actions"] = actions
            if comp:
                components[node.name] = comp
        results[name] = components
    with open(HWCOMPONENTS_JSON_PATH, "w") as f:
        json.dump(results, f, indent=2)
    print(f"Wrote {len(results)} arch results to {HWCOMPONENTS_JSON_PATH}")


if __name__ == "__main__":
    generate_hwcomponents()
    generate(fusion_choices=(True, False))
